@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2020 The Qt Company Ltd.
+** Copyright (C) 2021 The Qt Company Ltd.
 ** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtCore module of the Qt Toolkit.
@@ -43,6 +43,8 @@
 #include "qvariant.h"
 #include "qdatetime.h"
 
+#include "private/qstringiterator_p.h"
+#include "private/qgregoriancalendar_p.h"
 #ifdef Q_OS_DARWIN
 #include "private/qcore_mac_p.h"
 #include <CoreFoundation/CoreFoundation.h>
@@ -54,84 +56,218 @@ QT_BEGIN_NAMESPACE
 ** Wrappers for Mac locale system functions
 */
 
-static QByteArray envVarLocale()
-{
-    static QByteArray lang = 0;
-#ifdef Q_OS_UNIX
-    lang = qgetenv("LC_ALL");
-    if (lang.isEmpty())
-        lang = qgetenv("LC_NUMERIC");
-    if (lang.isEmpty())
-#endif
-        lang = qgetenv("LANG");
-    return lang;
-}
-
 static QString getMacLocaleName()
 {
-    QString result = QString::fromLocal8Bit(envVarLocale());
-
-    QString lang, script, cntry;
-    if (result.isEmpty()
-        || (result != QLatin1String("C") && !qt_splitLocaleName(result, lang, script, cntry))) {
-        QCFType<CFLocaleRef> l = CFLocaleCopyCurrent();
-        CFStringRef locale = CFLocaleGetIdentifier(l);
-        result = QString::fromCFString(locale);
-    }
-    return result;
+    QCFType<CFLocaleRef> l = CFLocaleCopyCurrent();
+    CFStringRef locale = CFLocaleGetIdentifier(l);
+    return QString::fromCFString(locale);
 }
 
-static QString macMonthName(int month, bool short_format)
+static QVariant macMonthName(int month, QSystemLocale::QueryType type)
 {
     month -= 1;
     if (month < 0 || month > 11)
-        return QString();
+        return {};
 
     QCFType<CFDateFormatterRef> formatter
         = CFDateFormatterCreate(0, QCFType<CFLocaleRef>(CFLocaleCopyCurrent()),
                                 kCFDateFormatterNoStyle,  kCFDateFormatterNoStyle);
+
+    CFDateFormatterKey formatterType;
+    switch (type) {
+        case QSystemLocale::MonthNameLong:
+            formatterType = kCFDateFormatterMonthSymbols;
+            break;
+        case QSystemLocale::MonthNameShort:
+            formatterType = kCFDateFormatterShortMonthSymbols;
+            break;
+        case QSystemLocale::MonthNameNarrow:
+            formatterType = kCFDateFormatterVeryShortMonthSymbols;
+            break;
+        case QSystemLocale::StandaloneMonthNameLong:
+            formatterType = kCFDateFormatterStandaloneMonthSymbols;
+            break;
+        case QSystemLocale::StandaloneMonthNameShort:
+            formatterType = kCFDateFormatterShortStandaloneMonthSymbols;
+            break;
+        case QSystemLocale::StandaloneMonthNameNarrow:
+            formatterType = kCFDateFormatterVeryShortStandaloneMonthSymbols;
+            break;
+        default:
+            qWarning("macMonthName: Unsupported query type %d", type);
+            return {};
+    }
     QCFType<CFArrayRef> values
-        = static_cast<CFArrayRef>(CFDateFormatterCopyProperty(formatter,
-                                  short_format ? kCFDateFormatterShortMonthSymbols
-                                               : kCFDateFormatterMonthSymbols));
+        = static_cast<CFArrayRef>(CFDateFormatterCopyProperty(formatter, formatterType));
+
     if (values != 0) {
         CFStringRef cfstring = static_cast<CFStringRef>(CFArrayGetValueAtIndex(values, month));
         return QString::fromCFString(cfstring);
     }
-    return QString();
+    return {};
 }
 
-static QString macDayName(int day, bool short_format)
+static QVariant macDayName(int day, QSystemLocale::QueryType type)
 {
     if (day < 1 || day > 7)
-        return QString();
+        return {};
 
     QCFType<CFDateFormatterRef> formatter
         = CFDateFormatterCreate(0, QCFType<CFLocaleRef>(CFLocaleCopyCurrent()),
                                 kCFDateFormatterNoStyle,  kCFDateFormatterNoStyle);
-    QCFType<CFArrayRef> values = static_cast<CFArrayRef>(CFDateFormatterCopyProperty(formatter,
-                                            short_format ? kCFDateFormatterShortWeekdaySymbols
-                                                         : kCFDateFormatterWeekdaySymbols));
+
+    CFDateFormatterKey formatterType;
+    switch (type) {
+    case QSystemLocale::DayNameLong:
+        formatterType = kCFDateFormatterWeekdaySymbols;
+        break;
+    case QSystemLocale::DayNameShort:
+        formatterType = kCFDateFormatterShortWeekdaySymbols;
+        break;
+    case QSystemLocale::DayNameNarrow:
+        formatterType = kCFDateFormatterVeryShortWeekdaySymbols;
+        break;
+    case QSystemLocale::StandaloneDayNameLong:
+        formatterType = kCFDateFormatterStandaloneWeekdaySymbols;
+        break;
+    case QSystemLocale::StandaloneDayNameShort:
+        formatterType = kCFDateFormatterShortStandaloneWeekdaySymbols;
+        break;
+    case QSystemLocale::StandaloneDayNameNarrow:
+        formatterType = kCFDateFormatterVeryShortStandaloneWeekdaySymbols;
+        break;
+    default:
+        qWarning("macDayName: Unsupported query type %d", type);
+        return {};
+    }
+    QCFType<CFArrayRef> values =
+            static_cast<CFArrayRef>(CFDateFormatterCopyProperty(formatter, formatterType));
+
     if (values != 0) {
         CFStringRef cfstring = static_cast<CFStringRef>(CFArrayGetValueAtIndex(values, day % 7));
         return QString::fromCFString(cfstring);
     }
-    return QString();
+    return {};
 }
 
-static QString macDateToString(const QDate &date, bool short_format)
+static QString macZeroDigit()
 {
-    QCFType<CFDateRef> myDate = QDateTime(date, QTime()).toCFDate();
-    QCFType<CFLocaleRef> mylocale = CFLocaleCopyCurrent();
-    CFDateFormatterStyle style = short_format ? kCFDateFormatterShortStyle : kCFDateFormatterLongStyle;
-    QCFType<CFDateFormatterRef> myFormatter
-        = CFDateFormatterCreate(kCFAllocatorDefault,
-                                mylocale, style,
-                                kCFDateFormatterNoStyle);
-    return QCFString(CFDateFormatterCreateStringWithDate(0, myFormatter, myDate));
+    QCFType<CFLocaleRef> locale = CFLocaleCopyCurrent();
+    QCFType<CFNumberFormatterRef> numberFormatter =
+            CFNumberFormatterCreate(nullptr, locale, kCFNumberFormatterNoStyle);
+    const int zeroDigit = 0;
+    QCFType<CFStringRef> value
+        = CFNumberFormatterCreateStringWithValue(nullptr, numberFormatter,
+                                                 kCFNumberIntType, &zeroDigit);
+    return QString::fromCFString(value);
 }
 
-static QString macTimeToString(const QTime &time, bool short_format)
+static QString zeroPad(QString &&number, qsizetype minDigits, const QString &zero)
+{
+    // Need to pad with zeros, possibly after a sign.
+    qsizetype insert = -1, digits = 0;
+    auto it = QStringIterator(number);
+    while (it.hasNext()) {
+        qsizetype here = it.index();
+        if (QChar::isDigit(it.next())) {
+            if (insert < 0)
+                insert = here;
+            ++digits;
+        } // else: assume we're stepping over a sign (or maybe grouping separator)
+    }
+    Q_ASSERT(digits > 0);
+    Q_ASSERT(insert >= 0);
+    while (digits++ < minDigits)
+        number.insert(insert, zero);
+
+    return std::move(number);
+}
+
+static QString trimTwoDigits(QString &&number)
+{
+    // Retain any sign, but remove all but the last two digits.
+    // We know number has at least four digits - it came from fourDigitYear().
+    // Note that each digit might be a surrogate pair.
+    qsizetype first = -1, prev = -1, last = -1;
+    auto it = QStringIterator(number);
+    while (it.hasNext()) {
+        qsizetype here = it.index();
+        if (QChar::isDigit(it.next())) {
+            if (first == -1)
+                last = first = here;
+            else if (last != -1)
+                prev = std::exchange(last, here);
+        }
+    }
+    Q_ASSERT(first >= 0);
+    Q_ASSERT(prev > first);
+    Q_ASSERT(last > prev);
+    number.remove(first, prev - first);
+    return std::move(number);
+}
+
+static QString fourDigitYear(int year, const QString &zero)
+{
+    // Return year formatted as an (at least) four digit number:
+    QCFType<CFLocaleRef> locale = CFLocaleCopyCurrent();
+    QCFType<CFNumberFormatterRef> numberFormatter =
+            CFNumberFormatterCreate(nullptr, locale, kCFNumberFormatterNoStyle);
+    QCFType<CFStringRef> value = CFNumberFormatterCreateStringWithValue(nullptr, numberFormatter,
+                                                                        kCFNumberIntType, &year);
+    auto text = QString::fromCFString(value);
+    if (year > -1000 && year < 1000)
+        text = zeroPad(std::move(text), 4, zero);
+    return text;
+}
+
+static QString macDateToStringImpl(QDate date, CFDateFormatterStyle style)
+{
+    QCFType<CFDateRef> myDate = date.startOfDay().toCFDate();
+    QCFType<CFLocaleRef> mylocale = CFLocaleCopyCurrent();
+    QCFType<CFDateFormatterRef> myFormatter
+        = CFDateFormatterCreate(kCFAllocatorDefault, mylocale, style,
+                                kCFDateFormatterNoStyle);
+    QCFType<CFStringRef> text = CFDateFormatterCreateStringWithDate(0, myFormatter, myDate);
+    return QString::fromCFString(text);
+}
+
+static QVariant macDateToString(QDate date, bool short_format)
+{
+    const int year = date.year();
+    QString fakeYear, trueYear;
+    if (year < 0) {
+        // System API (in macOS 11.0, at least) discards sign :-(
+        // Simply negating the year won't do as the resulting year typically has
+        // a different pattern of week-days.
+        int matcher = QGregorianCalendar::yearSharingWeekDays(date);
+        Q_ASSERT(matcher > 0);
+        Q_ASSERT(matcher % 100 != date.month());
+        Q_ASSERT(matcher % 100 != date.day());
+        // i.e. there can't be any confusion between the two-digit year and
+        // month or day-of-month in the formatted date.
+        QString zero = macZeroDigit();
+        fakeYear = fourDigitYear(matcher, zero);
+        trueYear = fourDigitYear(year, zero);
+        date = QDate(matcher, date.month(), date.day());
+    }
+    QString text = macDateToStringImpl(date, short_format
+                                       ? kCFDateFormatterShortStyle
+                                       : kCFDateFormatterLongStyle);
+    if (year < 0) {
+        if (text.contains(fakeYear))
+            return std::move(text).replace(fakeYear, trueYear);
+        // Cope with two-digit year:
+        fakeYear = trimTwoDigits(std::move(fakeYear));
+        trueYear = trimTwoDigits(std::move(trueYear));
+        if (text.contains(fakeYear))
+            return std::move(text).replace(fakeYear, trueYear);
+        // That should have worked.
+        qWarning("Failed to fix up year when formatting a date in year %d", year);
+    }
+    return text;
+}
+
+static QVariant macTimeToString(QTime time, bool short_format)
 {
     QCFType<CFDateRef> myDate = QDateTime(QDate::currentDate(), time).toCFDate();
     QCFType<CFLocaleRef> mylocale = CFLocaleCopyCurrent();
@@ -140,7 +276,8 @@ static QString macTimeToString(const QTime &time, bool short_format)
                                                                     mylocale,
                                                                     kCFDateFormatterNoStyle,
                                                                     style);
-    return QCFString(CFDateFormatterCreateStringWithDate(0, myFormatter, myDate));
+    QCFType<CFStringRef> text = CFDateFormatterCreateStringWithDate(0, myFormatter, myDate);
+    return QString::fromCFString(text);
 }
 
 // Mac uses the Unicode CLDR format codes
@@ -148,7 +285,7 @@ static QString macTimeToString(const QTime &time, bool short_format)
 // See also qtbase/util/locale_database/dateconverter.py
 // Makes the assumption that input formats are always well formed and consecutive letters
 // never exceed the maximum for the format code.
-static QString macToQtFormat(QStringView sys_fmt)
+static QVariant macToQtFormat(QStringView sys_fmt)
 {
     QString result;
     int i = 0;
@@ -164,7 +301,7 @@ static QString macToQtFormat(QStringView sys_fmt)
         }
 
         QChar c = sys_fmt.at(i);
-        int repeat = qt_repeatCount(sys_fmt.mid(i));
+        qsizetype repeat = qt_repeatCount(sys_fmt.mid(i));
 
         switch (c.unicode()) {
             // Qt does not support the following options
@@ -264,10 +401,10 @@ static QString macToQtFormat(QStringView sys_fmt)
         i += repeat;
     }
 
-    return result;
+    return !result.isEmpty() ? QVariant::fromValue(result) : QVariant();
 }
 
-QString getMacDateFormat(CFDateFormatterStyle style)
+static QVariant getMacDateFormat(CFDateFormatterStyle style)
 {
     QCFType<CFLocaleRef> l = CFLocaleCopyCurrent();
     QCFType<CFDateFormatterRef> formatter = CFDateFormatterCreate(kCFAllocatorDefault,
@@ -275,7 +412,7 @@ QString getMacDateFormat(CFDateFormatterStyle style)
     return macToQtFormat(QString::fromCFString(CFDateFormatterGetFormat(formatter)));
 }
 
-static QString getMacTimeFormat(CFDateFormatterStyle style)
+static QVariant getMacTimeFormat(CFDateFormatterStyle style)
 {
     QCFType<CFLocaleRef> l = CFLocaleCopyCurrent();
     QCFType<CFDateFormatterRef> formatter = CFDateFormatterCreate(kCFAllocatorDefault,
@@ -283,14 +420,16 @@ static QString getMacTimeFormat(CFDateFormatterStyle style)
     return macToQtFormat(QString::fromCFString(CFDateFormatterGetFormat(formatter)));
 }
 
-static QString getCFLocaleValue(CFStringRef key)
+static QVariant getCFLocaleValue(CFStringRef key)
 {
     QCFType<CFLocaleRef> locale = CFLocaleCopyCurrent();
     CFTypeRef value = CFLocaleGetValue(locale, key);
+    if (!value)
+        return QVariant();
     return QString::fromCFString(CFStringRef(static_cast<CFTypeRef>(value)));
 }
 
-static QLocale::MeasurementSystem macMeasurementSystem()
+static QVariant macMeasurementSystem()
 {
     QCFType<CFLocaleRef> locale = CFLocaleCopyCurrent();
     CFStringRef system = static_cast<CFStringRef>(CFLocaleGetValue(locale, kCFLocaleMeasurementSystem));
@@ -311,7 +450,7 @@ static quint8 macFirstDayOfWeek()
     return day;
 }
 
-static QString macCurrencySymbol(QLocale::CurrencySymbolFormat format)
+static QVariant macCurrencySymbol(QLocale::CurrencySymbolFormat format)
 {
     QCFType<CFLocaleRef> locale = CFLocaleCopyCurrent();
     switch (format) {
@@ -327,44 +466,33 @@ static QString macCurrencySymbol(QLocale::CurrencySymbolFormat format)
     default:
         break;
     }
-    return QString();
-}
-
-static QString macZeroDigit()
-{
-    QCFType<CFLocaleRef> locale = CFLocaleCopyCurrent();
-    QCFType<CFNumberFormatterRef> numberFormatter =
-            CFNumberFormatterCreate(nullptr, locale, kCFNumberFormatterNoStyle);
-    static const int zeroDigit = 0;
-    QCFType<CFStringRef> value = CFNumberFormatterCreateStringWithValue(nullptr, numberFormatter,
-                                                                        kCFNumberIntType, &zeroDigit);
-    return QString::fromCFString(value);
+    return {};
 }
 
 #ifndef QT_NO_SYSTEMLOCALE
-static QString macFormatCurrency(const QSystemLocale::CurrencyToStringArgument &arg)
+static QVariant macFormatCurrency(const QSystemLocale::CurrencyToStringArgument &arg)
 {
     QCFType<CFNumberRef> value;
-    switch (arg.value.type()) {
-    case QVariant::Int:
-    case QVariant::UInt: {
+    switch (arg.value.metaType().id()) {
+    case QMetaType::Int:
+    case QMetaType::UInt: {
         int v = arg.value.toInt();
         value = CFNumberCreate(NULL, kCFNumberIntType, &v);
         break;
     }
-    case QVariant::Double: {
+    case QMetaType::Double: {
         double v = arg.value.toDouble();
         value = CFNumberCreate(NULL, kCFNumberDoubleType, &v);
         break;
     }
-    case QVariant::LongLong:
-    case QVariant::ULongLong: {
+    case QMetaType::LongLong:
+    case QMetaType::ULongLong: {
         qint64 v = arg.value.toLongLong();
         value = CFNumberCreate(NULL, kCFNumberLongLongType, &v);
         break;
     }
     default:
-        return QString();
+        return {};
     }
 
     QCFType<CFLocaleRef> locale = CFLocaleCopyCurrent();
@@ -378,7 +506,7 @@ static QString macFormatCurrency(const QSystemLocale::CurrencyToStringArgument &
     return QString::fromCFString(result);
 }
 
-static QVariant macQuoteString(QSystemLocale::QueryType type, const QStringRef &str)
+static QVariant macQuoteString(QSystemLocale::QueryType type, QStringView str)
 {
     QString begin, end;
     QCFType<CFLocaleRef> locale = CFLocaleCopyCurrent();
@@ -400,25 +528,46 @@ static QVariant macQuoteString(QSystemLocale::QueryType type, const QStringRef &
 
 #ifndef QT_NO_SYSTEMLOCALE
 
-QLocale QSystemLocale::fallbackUiLocale() const
+QLocale QSystemLocale::fallbackLocale() const
 {
     return QLocale(getMacLocaleName());
 }
 
+template <auto CodeToValueFunction>
+static QVariant getLocaleValue(CFStringRef key)
+{
+    if (auto code = getCFLocaleValue(key); !code.isNull()) {
+        // If an invalid locale is requested with -AppleLocale, the system APIs
+        // will report invalid or empty locale values back to us, which codeToLanguage()
+        // and friends will fail to parse, resulting in returning QLocale::Any{L/C/S}.
+        // If this is the case, we fall down and return a null-variant, which
+        // QLocale's updateSystemPrivate() will interpret to use fallback logic.
+        if (auto value = CodeToValueFunction(code.toString()))
+            return value;
+    }
+    return QVariant();
+}
+
+static QLocale::Language codeToLanguage(QStringView s)
+{
+    return QLocalePrivate::codeToLanguage(s);
+}
+
 QVariant QSystemLocale::query(QueryType type, QVariant in) const
 {
     QMacAutoReleasePool pool;
+
     switch(type) {
-//     case Name:
-//         return getMacLocaleName();
-    case DecimalPoint: {
-        QString value = getCFLocaleValue(kCFLocaleDecimalSeparator);
-        return value.isEmpty() ? QVariant() : value;
-    }
-    case GroupSeparator: {
-        QString value = getCFLocaleValue(kCFLocaleGroupingSeparator);
-        return value.isEmpty() ? QVariant() : value;
-    }
+    case LanguageId:
+        return getLocaleValue<codeToLanguage>(kCFLocaleLanguageCode);
+    case TerritoryId:
+        return getLocaleValue<QLocalePrivate::codeToTerritory>(kCFLocaleCountryCode);
+    case ScriptId:
+        return getLocaleValue<QLocalePrivate::codeToScript>(kCFLocaleScriptCode);
+    case DecimalPoint:
+        return getCFLocaleValue(kCFLocaleDecimalSeparator);
+    case GroupSeparator:
+        return getCFLocaleValue(kCFLocaleGroupingSeparator);
     case DateFormatLong:
     case DateFormatShort:
         return getMacDateFormat(type == DateFormatShort
@@ -431,12 +580,18 @@ QVariant QSystemLocale::query(QueryType type, QVariant in) const
                                 : kCFDateFormatterLongStyle);
     case DayNameLong:
     case DayNameShort:
-        return macDayName(in.toInt(), (type == DayNameShort));
+    case DayNameNarrow:
+    case StandaloneDayNameLong:
+    case StandaloneDayNameShort:
+    case StandaloneDayNameNarrow:
+        return macDayName(in.toInt(), type);
     case MonthNameLong:
     case MonthNameShort:
+    case MonthNameNarrow:
     case StandaloneMonthNameLong:
     case StandaloneMonthNameShort:
-        return macMonthName(in.toInt(), (type == MonthNameShort || type == StandaloneMonthNameShort));
+    case StandaloneMonthNameNarrow:
+        return macMonthName(in.toInt(), type);
     case DateToStringShort:
     case DateToStringLong:
         return macDateToString(in.toDate(), (type == DateToStringShort));
@@ -448,10 +603,10 @@ QVariant QSystemLocale::query(QueryType type, QVariant in) const
     case PositiveSign:
         break;
     case ZeroDigit:
-        return QVariant(macZeroDigit());
+        return macZeroDigit();
 
     case MeasurementSystem:
-        return QVariant(static_cast<int>(macMeasurementSystem()));
+        return macMeasurementSystem();
 
     case AMText:
     case PMText: {
@@ -464,39 +619,24 @@ QVariant QSystemLocale::query(QueryType type, QVariant in) const
     case FirstDayOfWeek:
         return QVariant(macFirstDayOfWeek());
     case CurrencySymbol:
-        return QVariant(macCurrencySymbol(QLocale::CurrencySymbolFormat(in.toUInt())));
+        return macCurrencySymbol(QLocale::CurrencySymbolFormat(in.toUInt()));
     case CurrencyToString:
         return macFormatCurrency(in.value<QSystemLocale::CurrencyToStringArgument>());
     case UILanguages: {
-        QCFType<CFPropertyListRef> languages = CFPreferencesCopyValue(
-                 CFSTR("AppleLanguages"),
-                 kCFPreferencesAnyApplication,
-                 kCFPreferencesCurrentUser,
-                 kCFPreferencesAnyHost);
         QStringList result;
-        if (!languages)
-            return QVariant(result);
-
-        CFTypeID typeId = CFGetTypeID(languages);
-        if (typeId == CFArrayGetTypeID()) {
-            const int cnt = CFArrayGetCount(languages.as<CFArrayRef>());
-            result.reserve(cnt);
-            for (int i = 0; i < cnt; ++i) {
-                const QString lang = QString::fromCFString(
-                            static_cast<CFStringRef>(CFArrayGetValueAtIndex(languages.as<CFArrayRef>(), i)));
-                result.append(lang);
-            }
-        } else if (typeId == CFStringGetTypeID()) {
-            result = QStringList(QString::fromCFString(languages.as<CFStringRef>()));
-        } else {
-            qWarning("QLocale::uiLanguages(): CFPreferencesCopyValue returned unhandled type \"%ls\"; please report to http://bugreports.qt.io",
-                     qUtf16Printable(QString::fromCFString(CFCopyTypeIDDescription(typeId))));
+        QCFType<CFArrayRef> languages = CFLocaleCopyPreferredLanguages();
+        const int cnt = CFArrayGetCount(languages);
+        result.reserve(cnt);
+        for (int i = 0; i < cnt; ++i) {
+            const QString lang = QString::fromCFString(
+                static_cast<CFStringRef>(CFArrayGetValueAtIndex(languages, i)));
+            result.append(lang);
         }
         return QVariant(result);
     }
     case StringToStandardQuotation:
     case StringToAlternateQuotation:
-        return macQuoteString(type, in.value<QStringRef>());
+        return macQuoteString(type, in.value<QStringView>());
     default:
         break;
     }
