@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2020 The Qt Company Ltd.
+** Copyright (C) 2021 The Qt Company Ltd.
 ** Copyright (C) 2016 Intel Corporation.
 ** Contact: https://www.qt.io/licensing/
 **
@@ -46,11 +46,27 @@
 #include "qdatetime.h"
 #include "qdebug.h"
 
+#include "QtCore/private/qgregoriancalendar_p.h" // for yearSharingWeekDays()
+
 #ifdef Q_OS_WIN
 #   include <qt_windows.h>
 #   include <time.h>
 #endif
 
+#if QT_CONFIG(cpp_winrt) && !defined(Q_CC_CLANG)
+#   include <winrt/base.h>
+// Workaround for Windows SDK bug.
+// See https://github.com/microsoft/Windows.UI.Composition-Win32-Samples/issues/47
+namespace winrt::impl
+{
+    template <typename Async>
+    auto wait_for(Async const& async, Windows::Foundation::TimeSpan const& timeout);
+}
+#   include <winrt/Windows.Foundation.h>
+#   include <winrt/Windows.Foundation.Collections.h>
+#   include <winrt/Windows.System.UserProfile.h>
+#endif // QT_CONFIG(cpp_winrt) && !defined(Q_CC_CLANG)
+
 QT_BEGIN_NAMESPACE
 
 static QByteArray getWinLocaleName(LCID id = LOCALE_USER_DEFAULT);
@@ -94,6 +110,7 @@ struct QSystemLocalePrivate
     QVariant timeFormat(QLocale::FormatType);
     QVariant dateTimeFormat(QLocale::FormatType);
     QVariant dayName(int, QLocale::FormatType);
+    QVariant standaloneMonthName(int, QLocale::FormatType);
     QVariant monthName(int, QLocale::FormatType);
     QVariant toString(QDate, QLocale::FormatType);
     QVariant toString(QTime, QLocale::FormatType);
@@ -107,7 +124,7 @@ struct QSystemLocalePrivate
     QVariant toCurrencyString(const QSystemLocale::CurrencyToStringArgument &);
     QVariant uiLanguages();
     QVariant nativeLanguageName();
-    QVariant nativeCountryName();
+    QVariant nativeTerritoryName();
 
     void update();
 
@@ -134,6 +151,7 @@ private:
 
     SubstitutionType substitution();
     QString substituteDigits(QString &&string);
+    QString yearFix(int year, int fakeYear, QString &&formatted);
 
     static QString winToQtFormat(QStringView sys_fmt);
 
@@ -328,7 +346,7 @@ QVariant QSystemLocalePrivate::timeFormat(QLocale::FormatType type)
     case QLocale::NarrowFormat:
         break;
     }
-    return QVariant();
+    return {};
 }
 
 QVariant QSystemLocalePrivate::dateTimeFormat(QLocale::FormatType type)
@@ -366,7 +384,7 @@ QVariant QSystemLocalePrivate::dayName(int day, QLocale::FormatType type)
     return getLocaleInfo(short_day_map[day]);
 }
 
-QVariant QSystemLocalePrivate::monthName(int month, QLocale::FormatType type)
+QVariant QSystemLocalePrivate::standaloneMonthName(int month, QLocale::FormatType type)
 {
     static const LCTYPE short_month_map[]
         = { LOCALE_SABBREVMONTHNAME1, LOCALE_SABBREVMONTHNAME2, LOCALE_SABBREVMONTHNAME3,
@@ -380,38 +398,120 @@ QVariant QSystemLocalePrivate::monthName(int month, QLocale::FormatType type)
             LOCALE_SMONTHNAME7, LOCALE_SMONTHNAME8, LOCALE_SMONTHNAME9,
             LOCALE_SMONTHNAME10, LOCALE_SMONTHNAME11, LOCALE_SMONTHNAME12 };
 
-    month -= 1;
-    if (month < 0 || month > 11)
-        return QString();
+    if (month < 1 || month > 12)
+        return {};
 
-    LCTYPE lctype = (type == QLocale::ShortFormat || type == QLocale::NarrowFormat)
-            ? short_month_map[month] : long_month_map[month];
-    return getLocaleInfo(lctype);
+    // Month is Jan = 1, ... Dec = 12; adjust by 1 to match array indexing from 0:
+    return getLocaleInfo(
+        (type == QLocale::LongFormat ? long_month_map : short_month_map)[month - 1]);
+}
+
+QVariant QSystemLocalePrivate::monthName(int month, QLocale::FormatType type)
+{
+    SYSTEMTIME st = {};
+    st.wYear = 2001;
+    st.wMonth = month;
+    st.wDay = 10;
+
+    const DWORD flags{}; // Must be clear when passing a format string.
+    // MS's docs for the LOCALE_SMONTHNAME* say to include the day in a format.
+    // Educated guess: this works for the LOCALE_SABBREVMONTHNAME*, too, in so
+    // far as the abbreviated plain name might differ from abbreviated
+    // standalone one.
+    const wchar_t *const format = type == QLocale::LongFormat ? L"ddMMMM" : L"ddMMM";
+    wchar_t buf[255];
+    if (getDateFormat(flags, &st, format, buf, 255) > 2) {
+        // Elide the two digits of day number
+        QString text = QString::fromWCharArray(buf + 2);
+        if (substitution() == SAlways)
+            text = substituteDigits(std::move(text));
+        return text;
+    }
+    return {};
+}
+
+static QString fourDigitYear(int year)
+{
+    // Return year formatted as an (at least) four digit number:
+    return QStringLiteral("%1").arg(year, 4, 10, QChar(u'0'));
+}
+
+QString QSystemLocalePrivate::yearFix(int year, int fakeYear, QString &&formatted)
+{
+    // Replace our ersatz fakeYear (that MS formats faithfully) with the correct
+    // form of year.  We know the two-digit short form of fakeYear can not be
+    // mistaken for the month or day-of-month in the formatted date.
+    Q_ASSERT(fakeYear >= 1970 && fakeYear <= 2400);
+    const bool matchTwo = year >= 0 && year % 100 == fakeYear % 100;
+    auto yearUsed = fourDigitYear(fakeYear);
+    QString sign(year < 0 ? 1 : 0, QLatin1Char('-'));
+    auto trueYear = fourDigitYear(year < 0 ? -year : year);
+    if (formatted.contains(yearUsed))
+        return std::move(formatted).replace(yearUsed, sign + trueYear);
+
+    auto tail = QStringView{yearUsed}.last(2);
+    Q_ASSERT(!matchTwo || tail == QString(sign + trueYear.last(2)));
+    if (formatted.contains(tail)) {
+        if (matchTwo)
+            return std::move(formatted);
+        return std::move(formatted).replace(tail.toString(), sign + trueYear.last(2));
+    }
+
+    // Localized digits, perhaps ?
+    // First call to substituteDigits() ensures zero is initialized:
+    trueYear = substituteDigits(std::move(trueYear));
+    if (zero != u'0') {
+        yearUsed = substituteDigits(std::move(yearUsed));
+        if (year < 0)
+            sign = negativeSign().toString();
+
+        if (formatted.contains(yearUsed))
+            return std::move(formatted).replace(yearUsed, sign + trueYear);
+
+        const int twoDigits = 2 * zero.size();
+        tail = QStringView{yearUsed}.last(twoDigits);
+        if (formatted.contains(tail)) {
+            if (matchTwo)
+                return std::move(formatted);
+            return std::move(formatted).replace(tail.toString(), sign + trueYear.last(twoDigits));
+        }
+    }
+    qWarning("Failed to fix up year in formatted date-string using %d for %d", fakeYear, year);
+    return std::move(formatted);
 }
 
 QVariant QSystemLocalePrivate::toString(QDate date, QLocale::FormatType type)
 {
-    SYSTEMTIME st;
-    memset(&st, 0, sizeof(SYSTEMTIME));
-    st.wYear = date.year();
+    SYSTEMTIME st = {};
+    const int year = date.year();
+    // st.wYear is unsigned; and GetDateFormat() is documented to not handle
+    // dates before 1601.
+    const bool fixup = year < 1601;
+    st.wYear = fixup ? QGregorianCalendar::yearSharingWeekDays(date) : year;
     st.wMonth = date.month();
     st.wDay = date.day();
 
+    Q_ASSERT(!fixup || st.wYear % 100 != st.wMonth);
+    Q_ASSERT(!fixup || st.wYear % 100 != st.wDay);
+    // i.e. yearFix() can trust a match of its fakeYear's last two digits to not
+    // be the month or day part of the formatted date.
+
     DWORD flags = (type == QLocale::LongFormat ? DATE_LONGDATE : DATE_SHORTDATE);
     wchar_t buf[255];
     if (getDateFormat(flags, &st, NULL, buf, 255)) {
-        QString format = QString::fromWCharArray(buf);
+        QString text = QString::fromWCharArray(buf);
+        if (fixup)
+            text = yearFix(year, st.wYear, std::move(text));
         if (substitution() == SAlways)
-            format = substituteDigits(std::move(format));
-        return format;
+            text = substituteDigits(std::move(text));
+        return text;
     }
-    return QString();
+    return {};
 }
 
 QVariant QSystemLocalePrivate::toString(QTime time, QLocale::FormatType type)
 {
-    SYSTEMTIME st;
-    memset(&st, 0, sizeof(SYSTEMTIME));
+    SYSTEMTIME st = {};
     st.wHour = time.hour();
     st.wMinute = time.minute();
     st.wSecond = time.second();
@@ -424,12 +524,12 @@ QVariant QSystemLocalePrivate::toString(QTime time, QLocale::FormatType type)
 
     wchar_t buf[255];
     if (getTimeFormat(flags, &st, NULL, buf, 255)) {
-        QString format = QString::fromWCharArray(buf);
+        QString text = QString::fromWCharArray(buf);
         if (substitution() == SAlways)
-            format = substituteDigits(std::move(format));
-        return format;
+            text = substituteDigits(std::move(text));
+        return text;
     }
-    return QString();
+    return {};
 }
 
 QVariant QSystemLocalePrivate::toString(const QDateTime &dt, QLocale::FormatType type)
@@ -593,9 +693,20 @@ QVariant QSystemLocalePrivate::toCurrencyString(const QSystemLocale::CurrencyToS
 
 QVariant QSystemLocalePrivate::uiLanguages()
 {
+    QStringList result;
+#if QT_CONFIG(cpp_winrt) && !defined(Q_CC_CLANG)
+    using namespace winrt;
+    using namespace Windows::System::UserProfile;
+    auto languages = GlobalizationPreferences::Languages();
+    for (const auto &lang : languages)
+        result << QString::fromStdString(winrt::to_string(lang));
+    if (!result.isEmpty())
+        return result; // else just fall back to WIN32 API implementation
+#endif // QT_CONFIG(cpp_winrt) && !defined(Q_CC_CLANG)
+    // mingw and clang still have to use Win32 API
     unsigned long cnt = 0;
     QVarLengthArray<wchar_t, 64> buf(64);
-#  if !defined(QT_BOOTSTRAPPED) && !defined(QT_BUILD_QMAKE) // Not present in MinGW 4.9/bootstrap builds.
+#    if !defined(QT_BOOTSTRAPPED) // Not present in MinGW 4.9/bootstrap builds.
     unsigned long size = buf.size();
     if (!GetUserPreferredUILanguages(MUI_LANGUAGE_NAME, &cnt, buf.data(), &size)) {
         size = 0;
@@ -606,8 +717,7 @@ QVariant QSystemLocalePrivate::uiLanguages()
                 return QStringList();
         }
     }
-#  endif // !QT_BOOTSTRAPPED && !QT_BUILD_QMAKE
-    QStringList result;
+#    endif // !QT_BOOTSTRAPPED
     result.reserve(cnt);
     const wchar_t *str = buf.constData();
     for (; cnt > 0; --cnt) {
@@ -625,7 +735,7 @@ QVariant QSystemLocalePrivate::nativeLanguageName()
     return getLocaleInfo(LOCALE_SNATIVELANGUAGENAME);
 }
 
-QVariant QSystemLocalePrivate::nativeCountryName()
+QVariant QSystemLocalePrivate::nativeTerritoryName()
 {
     return getLocaleInfo(LOCALE_SNATIVECOUNTRYNAME);
 }
@@ -654,7 +764,7 @@ QString QSystemLocalePrivate::winToQtFormat(QStringView sys_fmt)
         }
 
         QChar c = sys_fmt.at(i);
-        int repeat = qt_repeatCount(sys_fmt.mid(i));
+        qsizetype repeat = qt_repeatCount(sys_fmt.mid(i));
 
         switch (c.unicode()) {
             // Date
@@ -702,7 +812,7 @@ QString QSystemLocalePrivate::winToQtFormat(QStringView sys_fmt)
     return result;
 }
 
-QLocale QSystemLocale::fallbackUiLocale() const
+QLocale QSystemLocale::fallbackLocale() const
 {
     return QLocale(QString::fromLatin1(getWinLocaleName()));
 }
@@ -735,12 +845,25 @@ QVariant QSystemLocale::query(QueryType type, QVariant in) const
         return d->dayName(in.toInt(), QLocale::LongFormat);
     case DayNameShort:
         return d->dayName(in.toInt(), QLocale::ShortFormat);
+    case DayNameNarrow:
+        return d->dayName(in.toInt(), QLocale::NarrowFormat);
+    case StandaloneDayNameLong:
+    case StandaloneDayNameShort:
+    case StandaloneDayNameNarrow:
+        // Windows does not provide standalone day names, so fall back to CLDR
+        return QVariant();
     case MonthNameLong:
-    case StandaloneMonthNameLong:
         return d->monthName(in.toInt(), QLocale::LongFormat);
+    case StandaloneMonthNameLong:
+        return d->standaloneMonthName(in.toInt(), QLocale::LongFormat);
     case MonthNameShort:
-    case StandaloneMonthNameShort:
         return d->monthName(in.toInt(), QLocale::ShortFormat);
+    case StandaloneMonthNameShort:
+        return d->standaloneMonthName(in.toInt(), QLocale::ShortFormat);
+    case MonthNameNarrow:
+    case StandaloneMonthNameNarrow:
+        // Windows provides no narrow month names, so we fall back to CLDR
+        return QVariant();
     case DateToStringShort:
         return d->toString(in.toDate(), QLocale::ShortFormat);
     case DateToStringLong:
@@ -757,13 +880,13 @@ QVariant QSystemLocale::query(QueryType type, QVariant in) const
         return d->zeroDigit();
     case LanguageId:
     case ScriptId:
-    case CountryId: {
+    case TerritoryId: {
         QLocaleId lid = QLocaleId::fromName(QString::fromLatin1(getWinLocaleName()));
         if (type == LanguageId)
             return lid.language_id;
         if (type == ScriptId)
-            return lid.script_id ? lid.script_id : ushort(fallbackUiLocale().script());
-        return lid.country_id ? lid.country_id : ushort(fallbackUiLocale().country());
+            return lid.script_id ? lid.script_id : ushort(fallbackLocale().script());
+        return lid.territory_id ? lid.territory_id : ushort(fallbackLocale().territory());
     }
     case MeasurementSystem:
         return d->measurementSystem();
@@ -786,8 +909,8 @@ QVariant QSystemLocale::query(QueryType type, QVariant in) const
         break;
     case NativeLanguageName:
         return d->nativeLanguageName();
-    case NativeCountryName:
-        return d->nativeCountryName();
+    case NativeTerritoryName:
+        return d->nativeTerritoryName();
     default:
         break;
     }
@@ -984,8 +1107,8 @@ static QString winIso639LangName(LCID id)
     if (!lang_code.isEmpty()) {
         const char *endptr;
         bool ok;
-        QByteArray latin1_lang_code = std::move(lang_code).toLatin1();
-        int i = qstrtoull(latin1_lang_code, &endptr, 16, &ok);
+        const QByteArray latin1 = std::move(lang_code).toLatin1();
+        const auto i = qstrntoull(latin1.data(), latin1.size(), &endptr, 16, &ok);
         if (ok && *endptr == '\0') {
             switch (i) {
                 case 0x814:
@@ -1027,7 +1150,7 @@ static QByteArray getWinLocaleName(LCID id)
         if (result == "C"
             || (!result.isEmpty() && qt_splitLocaleName(QString::fromLocal8Bit(result)))) {
             bool ok = false; // See if we have a Windows locale code instead of a locale name:
-            long id = qstrtoll(result.data(), 0, 0, &ok);
+            long id = qstrntoll(result.data(), result.size(), 0, 0, &ok);
             if (!ok || id == 0 || id < INT_MIN || id > INT_MAX) // Assume real locale name
                 return result;
             return winLangCodeToIsoName(int(id));
