@@ -40,6 +40,8 @@
 #include "qpointingdevice.h"
 #include "qpointingdevice_p.h"
 #include "qwindowsysteminterface_p.h"
+#include "qeventpoint_p.h"
+
 #include <QList>
 #include <QLoggingCategory>
 #include <QMutex>
@@ -177,9 +179,10 @@ QPointingDevice::QPointingDevice(QPointingDevicePrivate &d, QObject *parent)
 {
 }
 
+#if QT_DEPRECATED_SINCE(6, 0)
 /*!
     \internal
-    \deprecated Please use the constructor rather than setters.
+    \deprecated [6.0] Please use the constructor rather than setters.
 
     Sets the device type \a devType and infers the pointer type.
 */
@@ -210,7 +213,7 @@ void QPointingDevice::setType(DeviceType devType)
 
 /*!
     \internal
-    \deprecated Please use the constructor rather than setters.
+    \deprecated [6.0] Please use the constructor rather than setters.
 */
 void QPointingDevice::setCapabilities(QInputDevice::Capabilities caps)
 {
@@ -220,13 +223,14 @@ void QPointingDevice::setCapabilities(QInputDevice::Capabilities caps)
 
 /*!
     \internal
-    \deprecated Please use the constructor rather than setters.
+    \deprecated [6.0] Please use the constructor rather than setters.
 */
 void QPointingDevice::setMaximumTouchPoints(int c)
 {
     Q_D(QPointingDevice);
     d->maximumTouchPoints = c;
 }
+#endif // QT_DEPRECATED_SINCE(6, 0)
 
 /*!
     Returns the pointer type.
@@ -284,7 +288,7 @@ const QPointingDevice *QPointingDevice::primaryPointingDevice(const QString& sea
     const QPointingDevice *mouse = nullptr;
     const QPointingDevice *touchpad = nullptr;
     for (const QInputDevice *dev : v) {
-        if (dev->seatName() != seatName)
+        if (!seatName.isNull() && dev->seatName() != seatName)
             continue;
         if (dev->type() == QInputDevice::DeviceType::Mouse) {
             if (!mouse)
@@ -314,6 +318,9 @@ const QPointingDevice *QPointingDevice::primaryPointingDevice(const QString& sea
     return touchpad;
 }
 
+QPointingDevicePrivate::~QPointingDevicePrivate()
+    = default;
+
 /*!
     \internal
     Finds the device instance belonging to the drawing or eraser end of a particular stylus,
@@ -324,14 +331,16 @@ const QPointingDevice *QPointingDevice::primaryPointingDevice(const QString& sea
 
     If an instance matching the given \a deviceType and \a pointerType but with
     only a default-constructed \c uniqueId is found, it will be assumed to be
-    the one we're looking for, and its \c uniqueId will be updated to match the
-    given \a uniqueId. This is for the benefit of any platform plugin that can
+    the one we're looking for, its \c uniqueId will be updated to match the
+    given \a uniqueId, and its \c capabilities will be updated to match the
+    given \a capabilities. This is for the benefit of any platform plugin that can
     discover the tablet itself at startup, along with the supported stylus types,
     but then discovers specific styli later on as they come into proximity.
 */
 const QPointingDevice *QPointingDevicePrivate::queryTabletDevice(QInputDevice::DeviceType deviceType,
                                                                  QPointingDevice::PointerType pointerType,
                                                                  QPointingDeviceUniqueId uniqueId,
+                                                                 QPointingDevice::Capabilities capabilities,
                                                                  qint64 systemId)
 {
     const auto &devices = QInputDevice::devices();
@@ -346,7 +355,9 @@ const QPointingDevice *QPointingDevicePrivate::queryTabletDevice(QInputDevice::D
                 (devPriv->uniqueId == uniqueId || uniqueIdDiscovered)) {
             if (uniqueIdDiscovered) {
                 const_cast<QPointingDevicePrivate *>(devPriv)->uniqueId = uniqueId;
-                qCDebug(lcQpaInputDevices) << "discovered unique ID of tablet tool" << pdev;
+                if (capabilities)
+                    const_cast<QPointingDevicePrivate *>(devPriv)->capabilities = capabilities;
+                qCDebug(lcQpaInputDevices) << "discovered unique ID and capabilities of tablet tool" << pdev;
             }
             return pdev;
         }
@@ -418,13 +429,12 @@ QPointingDevicePrivate::EventPointData *QPointingDevicePrivate::queryPointById(i
 */
 QPointingDevicePrivate::EventPointData *QPointingDevicePrivate::pointById(int id) const
 {
-    auto it = activePoints.find(id);
-    if (it == activePoints.end()) {
+    const auto [it, inserted] = activePoints.try_emplace(id);
+    if (inserted) {
         Q_Q(const QPointingDevice);
-        QPointingDevicePrivate::EventPointData epd;
-        QMutableEventPoint::from(epd.eventPoint).setId(id);
-        QMutableEventPoint::from(epd.eventPoint).setDevice(q);
-        return &activePoints.insert(id, epd).first.value();
+        auto &epd = it.value();
+        QMutableEventPoint::setId(epd.eventPoint, id);
+        QMutableEventPoint::setDevice(epd.eventPoint, q);
     }
     return &it.value();
 }
@@ -447,7 +457,7 @@ void QPointingDevicePrivate::removePointById(int id)
 QObject *QPointingDevicePrivate::firstActiveTarget() const
 {
     for (auto &pt : activePoints.values()) {
-        if (auto target = QMutableEventPoint::constFrom(pt.eventPoint).target())
+        if (auto target = QMutableEventPoint::target(pt.eventPoint))
             return target;
     }
     return nullptr;
@@ -462,7 +472,7 @@ QObject *QPointingDevicePrivate::firstActiveTarget() const
 QWindow *QPointingDevicePrivate::firstActiveWindow() const
 {
     for (auto &pt : activePoints.values()) {
-        if (auto window = QMutableEventPoint::constFrom(pt.eventPoint).window())
+        if (auto window = QMutableEventPoint::window(pt.eventPoint))
             return window;
     }
     return nullptr;
@@ -501,7 +511,7 @@ void QPointingDevicePrivate::setExclusiveGrabber(const QPointerEvent *event, con
                                << "@" << point.scenePosition()
                                << ": grab" << oldGrabber << "->" << exclusiveGrabber;
     }
-    QMutableEventPoint::from(persistentPoint->eventPoint).setGlobalGrabPosition(point.globalPosition());
+    QMutableEventPoint::setGlobalGrabPosition(persistentPoint->eventPoint, point.globalPosition());
     if (exclusiveGrabber)
         emit q->grabChanged(exclusiveGrabber, QPointingDevice::GrabExclusive, event, point);
     else
@@ -800,3 +810,5 @@ size_t qHash(QPointingDeviceUniqueId key, size_t seed) noexcept
 }
 
 QT_END_NAMESPACE
+
+#include "moc_qpointingdevice.cpp"
