@@ -40,7 +40,6 @@
 #include "qwindowsfontdatabasebase_p.h"
 #include "qwindowsfontdatabase_p.h"
 
-#include <QtCore/private/qsystemlibrary_p.h>
 #include <QtCore/QThreadStorage>
 #include <QtCore/QtEndian>
 
@@ -582,36 +581,19 @@ bool QWindowsFontDatabaseBase::init(QSharedPointer<QWindowsFontEngineData> d)
 }
 
 #if QT_CONFIG(directwrite) && QT_CONFIG(direct2d)
-// ### Qt 6: Link directly to dwrite instead
-typedef HRESULT (WINAPI *DWriteCreateFactoryType)(DWRITE_FACTORY_TYPE, const IID &, IUnknown **);
-static inline DWriteCreateFactoryType resolveDWriteCreateFactory()
-{
-    QSystemLibrary library(QStringLiteral("dwrite"));
-    QFunctionPointer result = library.resolve("DWriteCreateFactory");
-    if (Q_UNLIKELY(!result)) {
-        qWarning("Unable to load dwrite.dll");
-        return nullptr;
-    }
-    return reinterpret_cast<DWriteCreateFactoryType>(result);
-}
-
 void QWindowsFontDatabaseBase::createDirectWriteFactory(IDWriteFactory **factory)
 {
     *factory = nullptr;
-
-    static const DWriteCreateFactoryType dWriteCreateFactory = resolveDWriteCreateFactory();
-    if (!dWriteCreateFactory)
-        return;
-
     IUnknown *result = nullptr;
+
 #  if QT_CONFIG(directwrite3)
-    dWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, __uuidof(IDWriteFactory3), &result);
+    DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, __uuidof(IDWriteFactory3), &result);
 #  endif
     if (result == nullptr)
-        dWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, __uuidof(IDWriteFactory2), &result);
+        DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, __uuidof(IDWriteFactory2), &result);
 
     if (result == nullptr) {
-        if (FAILED(dWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, __uuidof(IDWriteFactory), &result))) {
+        if (FAILED(DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, __uuidof(IDWriteFactory), &result))) {
             qErrnoWarning("DWriteCreateFactory failed");
             return;
         }
@@ -713,16 +695,6 @@ LOGFONT QWindowsFontDatabaseBase::fontDefToLOGFONT(const QFontDef &request, cons
         fam.truncate(LF_FACESIZE - 1);
     }
 
-    if (fam.isEmpty())
-        fam = QStringLiteral("MS Sans Serif");
-
-    if (fam == QLatin1String("MS Sans Serif")
-        && (request.style == QFont::StyleItalic || (-lf.lfHeight > 18 && -lf.lfHeight != 24))) {
-        fam = QStringLiteral("Arial"); // MS Sans Serif has bearing problems in italic, and does not scale
-    }
-    if (fam == QLatin1String("Courier") && !(request.styleStrategy & QFont::PreferBitmap))
-        fam = QStringLiteral("Courier New");
-
     memcpy(lf.lfFaceName, fam.utf16(), fam.size() * sizeof(wchar_t));
 
     return lf;
@@ -964,4 +936,23 @@ QStringList QWindowsFontDatabaseBase::extraTryFontsForFamily(const QString &fami
     return result;
 }
 
+QFontDef QWindowsFontDatabaseBase::sanitizeRequest(QFontDef request) const
+{
+    QFontDef req = request;
+    const QString fam = request.families.front();
+    if (fam.isEmpty())
+        req.families[0] = QStringLiteral("MS Sans Serif");
+
+    if (fam == QLatin1String("MS Sans Serif")) {
+        int height = -qRound(request.pixelSize);
+        // MS Sans Serif has bearing problems in italic, and does not scale
+        if (request.style == QFont::StyleItalic || (height > 18 && height != 24))
+            req.families[0] = QStringLiteral("Arial");
+    }
+
+    if (!(request.styleStrategy & QFont::StyleStrategy::PreferBitmap) && fam == u"Courier")
+        req.families[0] = QStringLiteral("Courier New");
+    return req;
+}
+
 QT_END_NAMESPACE
