@@ -97,22 +97,7 @@ QT_BEGIN_NAMESPACE
     calling QWindow::setVulkanInstance(). Thus a typical application pattern is
     the following:
 
-    \code
-    int main(int argc, char **argv)
-    {
-        QGuiApplication app(argc, argv);
-
-        QVulkanInstance inst;
-        if (!inst.create())
-            return 1;
-
-        ...
-        window->setVulkanInstance(&inst);
-        window->show();
-
-        return app.exec();
-    }
-    \endcode
+    \snippet code/src_gui_vulkan_qvulkaninstance.cpp 0
 
     \section1 Configuration
 
@@ -138,31 +123,12 @@ QT_BEGIN_NAMESPACE
     For example, to enable the standard validation layers, one could do the
     following:
 
-    \code
-        QVulkanInstance inst;
-
-        // Enable validation layer, if supported. Messages go to qDebug by default.
-        inst.setLayers(QByteArrayList() << "VK_LAYER_LUNARG_standard_validation");
-
-        bool ok = inst.create();
-        if (!ok)
-            ... // Vulkan not available
-        if (!inst.layers().contains("VK_LAYER_LUNARG_standard_validation"))
-            ... // validation layer not available
-    \endcode
+    \snippet code/src_gui_vulkan_qvulkaninstance.cpp 1
 
     Or, alternatively, to make decisions before attempting to create a Vulkan
     instance:
 
-    \code
-        QVulkanInstance inst;
-
-        if (inst.supportedLayers().contains("VK_LAYER_LUNARG_standard_validation"))
-            ...
-
-        bool ok = inst.create();
-        ...
-    \endcode
+    \snippet code/src_gui_vulkan_qvulkaninstance.cpp 2
 
     \section1 Adopting an Existing Instance
 
@@ -188,9 +154,9 @@ QT_BEGIN_NAMESPACE
 
     \note QVulkanFunctions and QVulkanDeviceFunctions are generated from the
     Vulkan API XML specifications when building the Qt libraries. Therefore no
-    documentation is provided for them. They contain the Vulkan 1.0 functions
+    documentation is provided for them. They contain the Vulkan 1.2 functions
     with the same signatures as described in the
-    \l{https://www.khronos.org/registry/vulkan/specs/1.0/html/}{Vulkan API
+    \l{https://www.khronos.org/registry/vulkan/specs/1.2/html/}{Vulkan API
     documentation}.
 
     \section1 Getting a Native Vulkan Surface for a Window
@@ -229,62 +195,7 @@ QT_BEGIN_NAMESPACE
 
     The following is the basic outline of creating a Vulkan-capable QWindow:
 
-    \code
-    class VulkanWindow : public QWindow
-    {
-    public:
-        VulkanWindow() {
-            setSurfaceType(VulkanSurface);
-        }
-
-        void exposeEvent(QExposeEvent *) {
-            if (isExposed()) {
-                if (!m_initialized) {
-                    m_initialized = true;
-                    // initialize device, swapchain, etc.
-                    QVulkanInstance *inst = vulkanInstance();
-                    QVulkanFunctions *f = inst->functions();
-                    uint32_t devCount = 0;
-                    f->vkEnumeratePhysicalDevices(inst->vkInstance(), &devCount, nullptr);
-                    ...
-                    // build the first frame
-                    render();
-                }
-            }
-        }
-
-        bool event(QEvent *e) {
-            if (e->type == QEvent::UpdateRequest)
-                render();
-            return QWindow::event(e);
-        }
-
-        void render() {
-           ...
-           requestUpdate(); // render continuously
-        }
-
-    private:
-        bool m_initialized = false;
-    };
-
-    int main(int argc, char **argv)
-    {
-        QGuiApplication app(argc, argv);
-
-        QVulkanInstance inst;
-        if (!inst.create()) {
-            qWarning("Vulkan not available");
-            return 1;
-        }
-
-        VulkanWindow window;
-        window.showMaximized();
-
-        return app.exec();
-
-    }
-    \endcode
+    \snippet code/src_gui_vulkan_qvulkaninstance.cpp 3
 
     \note In addition to expose, a well-behaving window implementation will
     also have to take care of additional events like resize and
@@ -340,7 +251,6 @@ public:
     QVulkanInstancePrivate(QVulkanInstance *q)
         : q_ptr(q),
           vkInst(VK_NULL_HANDLE),
-          flags(0),
           errorCode(VK_SUCCESS)
     { }
     ~QVulkanInstancePrivate() { reset(); }
@@ -358,6 +268,7 @@ public:
     VkResult errorCode;
     QScopedPointer<QVulkanFunctions> funcs;
     QHash<VkDevice, QVulkanDeviceFunctions *> deviceFuncs;
+    QList<QVulkanInstance::DebugFilter> debugFilters;
 };
 
 bool QVulkanInstancePrivate::ensureVulkan()
@@ -395,7 +306,7 @@ QVulkanInstance::QVulkanInstance()
 /*!
     Destructor.
 
-    \note current() will return \c nullptr once the instance is destroyed.
+    \note current() will return \nullptr once the instance is destroyed.
  */
 QVulkanInstance::~QVulkanInstance()
 {
@@ -404,6 +315,7 @@ QVulkanInstance::~QVulkanInstance()
 
 /*!
     \class QVulkanLayer
+    \inmodule QtGui
     \brief Represents information about a Vulkan layer.
  */
 
@@ -447,7 +359,7 @@ QVulkanInstance::~QVulkanInstance()
 */
 
 /*!
-    \fn uint qHash(const QVulkanLayer &key, uint seed)
+    \fn size_t qHash(const QVulkanLayer &key, size_t seed)
     \since 5.10
     \relates QVulkanLayer
 
@@ -457,6 +369,7 @@ QVulkanInstance::~QVulkanInstance()
 
 /*!
     \class QVulkanExtension
+    \inmodule QtGui
     \brief Represents information about a Vulkan extension.
  */
 
@@ -490,7 +403,7 @@ QVulkanInstance::~QVulkanInstance()
 */
 
 /*!
-    \fn uint qHash(const QVulkanExtension &key, uint seed)
+    \fn size_t qHash(const QVulkanExtension &key, size_t seed)
     \since 5.10
     \relates QVulkanExtension
 
@@ -500,42 +413,75 @@ QVulkanInstance::~QVulkanInstance()
 
 /*!
     \class QVulkanInfoVector
-    \brief A specialized QVector for QVulkanLayer and QVulkanExtension.
+    \inmodule QtGui
+    \brief A specialized QList for QVulkanLayer and QVulkanExtension.
  */
 
 /*!
-    \fn bool QVulkanInfoVector::contains(const QByteArray &name) const
+    \fn template<typename T> bool QVulkanInfoVector<T>::contains(const QByteArray &name) const
 
-    \return true if the vector contains a layer or extension with the given \a name.
+    \return true if the list contains a layer or extension with the given \a name.
  */
 
 /*!
-    \fn bool QVulkanInfoVector::contains(const QByteArray &name, int minVersion) const
+    \fn template<typename T> bool QVulkanInfoVector<T>::contains(const QByteArray &name, int minVersion) const
 
-    \return true if the vector contains a layer or extension with the given
+    \return true if the list contains a layer or extension with the given
     \a name and a version same as or newer than \a minVersion.
  */
 
 /*!
+    \fn QVulkanInfoVector<QVulkanLayer> QVulkanInstance::supportedLayers() const
     \return the list of supported instance-level layers.
 
     \note This function can be called before create().
  */
+
+/*!
+    \internal
+ */
 QVulkanInfoVector<QVulkanLayer> QVulkanInstance::supportedLayers()
 {
     return d_ptr->ensureVulkan() ? d_ptr->platformInst->supportedLayers() : QVulkanInfoVector<QVulkanLayer>();
 }
 
 /*!
+    \fn QVulkanInfoVector<QVulkanExtension> QVulkanInstance::supportedExtensions() const
     \return the list of supported instance-level extensions.
 
     \note This function can be called before create().
  */
+
+/*!
+    \internal
+ */
 QVulkanInfoVector<QVulkanExtension> QVulkanInstance::supportedExtensions()
 {
     return d_ptr->ensureVulkan() ? d_ptr->platformInst->supportedExtensions() : QVulkanInfoVector<QVulkanExtension>();
 }
 
+/*!
+    \return the version of instance-level functionality supported by the Vulkan
+    implementation.
+
+    In practice this is either the value returned from
+    vkEnumerateInstanceVersion, if that function is available (with Vulkan 1.1
+    and newer), or 1.0.
+
+    Applications that want to branch in their Vulkan feature and API usage
+    based on what Vulkan version is available at run time, can use this function
+    to determine what version to pass in to setApiVersion() before calling
+    create().
+
+    \note This function can be called before create().
+
+    \sa setApiVersion()
+ */
+QVersionNumber QVulkanInstance::supportedApiVersion() const
+{
+    return d_ptr->ensureVulkan() ? d_ptr->platformInst->supportedApiVersion() : QVersionNumber();
+}
+
 /*!
     Makes QVulkanInstance adopt an existing VkInstance handle instead of
     creating a new one.
@@ -613,13 +559,26 @@ void QVulkanInstance::setExtensions(const QByteArrayList &extensions)
 }
 
 /*!
-    Specifies the Vulkan API against which the application expects to run.
+    Specifies the highest Vulkan API version the application is designed to use.
 
-    By default no \a vulkanVersion is specified, and so no version check is performed
-    during Vulkan instance creation.
+    By default \a vulkanVersion is 0, which maps to Vulkan 1.0.
 
     \note This function can only be called before create() and has no effect if
     called afterwards.
+
+    \note Be aware that Vulkan 1.1 changes the behavior with regards to the
+    Vulkan API version field. In Vulkan 1.0 specifying an unsupported \a
+    vulkanVersion led to failing create() with \c VK_ERROR_INCOMPATIBLE_DRIVER,
+    as was mandated by the specification. Starting with Vulkan 1.1, the
+    specification disallows this, the driver must accept any version without
+    failing the instance creation.
+
+    Application developers are advised to familiarize themselves with the \c
+    apiVersion notes in
+    \l{https://www.khronos.org/registry/vulkan/specs/1.2-extensions/man/html/VkApplicationInfo.html}{the
+    Vulkan specification}.
+
+    \sa supportedApiVersion()
  */
 void QVulkanInstance::setApiVersion(const QVersionNumber &vulkanVersion)
 {
@@ -659,6 +618,7 @@ bool QVulkanInstance::create()
         d_ptr->extensions = d_ptr->platformInst->enabledExtensions();
         d_ptr->errorCode = VK_SUCCESS;
         d_ptr->funcs.reset(new QVulkanFunctions(this));
+        d_ptr->platformInst->setDebugFilters(d_ptr->debugFilters);
         return true;
     }
 
@@ -703,7 +663,7 @@ VkResult QVulkanInstance::errorCode() const
 }
 
 /*!
-    \return the VkInstance handle this QVulkanInstance wraps, or \c null if
+    \return the VkInstance handle this QVulkanInstance wraps, or \nullptr if
     create() has not yet been successfully called and no existing instance has
     been provided via setVkInstance().
  */
@@ -779,7 +739,17 @@ QPlatformVulkanInstance *QVulkanInstance::handle() const
     \note The returned object is owned and managed by the QVulkanInstance. Do
     not destroy or alter it.
 
-    \sa deviceFunctions()
+    The functions from the core Vulkan 1.0 API will be available always. When it
+    comes to higher Vulkan versions, such as, 1.1 and 1.2, the QVulkanFunctions
+    object will try to resolve the core API functions for those as well, but if
+    the Vulkan instance implementation at run time has no support for those,
+    calling any such unsupported function will lead to unspecified behavior. In
+    addition, to properly enable support for Vulkan versions higher than 1.0, an
+    appropriate instance API version may need to be set by calling
+    setApiVersion() before create(). To query the Vulkan implementation's
+    instance-level version, call supportedApiVersion().
+
+    \sa deviceFunctions(), supportedApiVersion()
  */
 QVulkanFunctions *QVulkanInstance::functions() const
 {
@@ -804,6 +774,16 @@ QVulkanFunctions *QVulkanInstance::functions() const
     to the application to notify the QVulkanInstance by calling
     resetDeviceFunctions().
 
+    The functions from the core Vulkan 1.0 API will be available always. When
+    it comes to higher Vulkan versions, such as, 1.1 and 1.2, the
+    QVulkanDeviceFunctions object will try to resolve the core API functions
+    for those as well, but if the Vulkan physical device at run time has no
+    support for those, calling any such unsupported function will lead to
+    unspecified behavior. To properly enable support for Vulkan versions higher
+    than 1.0, an appropriate instance API version may need to be set by calling
+    setApiVersion() before create(). In addition, applications are expected to
+    check the physical device's apiVersion in VkPhysicalDeviceProperties.
+
     \sa functions(), resetDeviceFunctions()
  */
 QVulkanDeviceFunctions *QVulkanInstance::deviceFunctions(VkDevice device)
@@ -846,7 +826,7 @@ VkSurfaceKHR QVulkanInstance::surfaceForWindow(QWindow *window)
     // VkSurfaceKHR is non-dispatchable and maps to a pointer on x64 and a uint64 on x86.
     // Therefore a pointer is returned from the platform plugin, not the value itself.
     void *p = nativeInterface->nativeResourceForWindow(QByteArrayLiteral("vkSurface"), window);
-    return p ? *static_cast<VkSurfaceKHR *>(p) : 0;
+    return p ? *static_cast<VkSurfaceKHR *>(p) : VK_NULL_HANDLE;
 }
 
 /*!
@@ -861,6 +841,22 @@ bool QVulkanInstance::supportsPresent(VkPhysicalDevice physicalDevice, uint32_t
     return d_ptr->platformInst->supportsPresent(physicalDevice, queueFamilyIndex, window);
 }
 
+/*!
+    This function should be called by the application's renderer before queuing
+    a present operation for \a window.
+
+    While on some platforms this will be a no-op, some may perform windowing
+    system dependent synchronization. For example, on Wayland this will
+    add send a wl_surface.frame request in order to prevent the driver from
+    blocking for minimized windows.
+
+    \since 5.15
+ */
+void QVulkanInstance::presentAboutToBeQueued(QWindow *window)
+{
+    d_ptr->platformInst->presentAboutToBeQueued(window);
+}
+
 /*!
     This function should be called by the application's renderer after queuing
     a present operation for \a window.
@@ -874,6 +870,50 @@ void QVulkanInstance::presentQueued(QWindow *window)
     d_ptr->platformInst->presentQueued(window);
 }
 
+/*!
+    \typedef QVulkanInstance::DebugFilter
+
+    Typedef for debug filtering callback functions.
+
+    \sa installDebugOutputFilter(), removeDebugOutputFilter()
+ */
+
+/*!
+    Installs a \a filter function that is called for every Vulkan debug
+    message. When the callback returns \c true, the message is stopped (filtered
+    out) and will not appear on the debug output.
+
+    \note Filtering is only effective when NoDebugOutputRedirect is not
+    \l{setFlags()}{set}. Installing filters has no effect otherwise.
+
+    \note This function can be called before create().
+
+    \sa removeDebugOutputFilter()
+ */
+void QVulkanInstance::installDebugOutputFilter(DebugFilter filter)
+{
+    if (!d_ptr->debugFilters.contains(filter)) {
+        d_ptr->debugFilters.append(filter);
+        if (d_ptr->platformInst)
+            d_ptr->platformInst->setDebugFilters(d_ptr->debugFilters);
+    }
+}
+
+/*!
+    Removes a \a filter function previously installed by
+    installDebugOutputFilter().
+
+    \note This function can be called before create().
+
+    \sa installDebugOutputFilter()
+ */
+void QVulkanInstance::removeDebugOutputFilter(DebugFilter filter)
+{
+    d_ptr->debugFilters.removeOne(filter);
+    if (d_ptr->platformInst)
+        d_ptr->platformInst->setDebugFilters(d_ptr->debugFilters);
+}
+
 #ifndef QT_NO_DEBUG_STREAM
 QDebug operator<<(QDebug dbg, const QVulkanLayer &layer)
 {
