@@ -199,7 +199,7 @@ QNetworkReplyHttpImpl::QNetworkReplyHttpImpl(QNetworkAccessManager* const manage
         if (d->synchronous && outgoingData) {
             // The synchronous HTTP is a corner case, we will put all upload data in one big QByteArray in the outgoingDataBuffer.
             // Yes, this is not the most efficient thing to do, but on the other hand synchronous XHR needs to die anyway.
-            d->outgoingDataBuffer = QSharedPointer<QRingBuffer>::create();
+            d->outgoingDataBuffer = std::make_shared<QRingBuffer>();
             qint64 previousDataSize = 0;
             do {
                 previousDataSize = d->outgoingDataBuffer->size();
@@ -305,6 +305,13 @@ qint64 QNetworkReplyHttpImpl::bytesAvailable() const
         return QNetworkReply::bytesAvailable() + d->downloadBufferCurrentSize - d->downloadBufferReadPosition;
     }
 
+    if (d->decompressHelper.isValid()) {
+        if (d->decompressHelper.isCountingBytes())
+            return QNetworkReply::bytesAvailable() + d->decompressHelper.uncompressedSize();
+        if (d->decompressHelper.hasData())
+            return QNetworkReply::bytesAvailable() + 1;
+    }
+
     // normal buffer
     return QNetworkReply::bytesAvailable();
 }
@@ -345,6 +352,31 @@ qint64 QNetworkReplyHttpImpl::readData(char* data, qint64 maxlen)
 
     }
 
+    if (d->decompressHelper.isValid() && (d->decompressHelper.hasData() || !isFinished())) {
+        if (maxlen == 0 || !d->decompressHelper.hasData())
+            return 0;
+        const qint64 bytesRead = d->decompressHelper.read(data, maxlen);
+        if (!d->decompressHelper.isValid()) {
+            d->error(QNetworkReplyImpl::NetworkError::UnknownContentError,
+                     QCoreApplication::translate("QHttp", "Decompression failed: %1")
+                             .arg(d->decompressHelper.errorString()));
+            return -1;
+        }
+        if (d->cacheSaveDevice) {
+            // Need to write to the cache now that we have the data
+            d->cacheSaveDevice->write(data, bytesRead);
+            // ... and if we've read everything then the cache can be closed.
+            if (isFinished() && !d->decompressHelper.hasData())
+                d->completeCacheSave();
+        }
+        // In case of buffer size restriction we need to emit that it has been emptied
+        qint64 wasBuffered = d->bytesBuffered;
+        d->bytesBuffered = 0;
+        if (readBufferSize())
+            emit readBufferFreed(wasBuffered);
+        return bytesRead;
+    }
+
     // normal buffer
     if (d->state == d->Finished || d->state == d->Aborted)
         return -1;
@@ -447,8 +479,8 @@ QNetworkReplyHttpImplPrivate::QNetworkReplyHttpImplPrivate()
     , downloadBufferReadPosition(0)
     , downloadBufferCurrentSize(0)
     , downloadZerocopyBuffer(nullptr)
-    , pendingDownloadDataEmissions(QSharedPointer<QAtomicInt>::create())
-    , pendingDownloadProgressEmissions(QSharedPointer<QAtomicInt>::create())
+    , pendingDownloadDataEmissions(std::make_shared<QAtomicInt>())
+    , pendingDownloadProgressEmissions(std::make_shared<QAtomicInt>())
     #ifndef QT_NO_SSL
     , pendingIgnoreAllSslErrors(false)
     #endif
@@ -680,6 +712,7 @@ void QNetworkReplyHttpImplPrivate::postRequest(const QNetworkRequest &newHttpReq
     httpRequest.setRedirectPolicy(redirectPolicy);
 
     httpRequest.setPriority(convert(newHttpRequest.priority()));
+    loadingFromCache = false;
 
     switch (operation) {
     case QNetworkAccessManager::GetOperation:
@@ -759,6 +792,12 @@ void QNetworkReplyHttpImplPrivate::postRequest(const QNetworkRequest &newHttpReq
         allowed.isValid() && allowed.canConvert<bool>()) {
         httpRequest.setHTTP2Allowed(allowed.value<bool>());
     }
+    auto h2cAttribute = request.attribute(QNetworkRequest::Http2CleartextAllowedAttribute);
+    // ### Qt7: Stop checking the environment variable
+    if (h2cAttribute.toBool()
+        || (!h2cAttribute.isValid() && qEnvironmentVariableIsSet("QT_NETWORK_H2C_ALLOWED"))) {
+        httpRequest.setH2cAllowed(true);
+    }
 
     if (request.attribute(QNetworkRequest::Http2DirectAttribute).toBool()) {
         // Intentionally mutually exclusive - cannot be both direct and 'allowed'
@@ -774,7 +813,6 @@ void QNetworkReplyHttpImplPrivate::postRequest(const QNetworkRequest &newHttpReq
     if (request.attribute(QNetworkRequest::EmitAllUploadProgressSignalsAttribute).toBool())
         emitAllUploadProgressSignals = true;
 
-    httpRequest.setMinimumArchiveBombSize(newHttpRequest.decompressedSafetyCheckThreshold());
     httpRequest.setPeerVerifyName(newHttpRequest.peerVerifyName());
 
     // Create the HTTP thread delegate
@@ -782,6 +820,9 @@ void QNetworkReplyHttpImplPrivate::postRequest(const QNetworkRequest &newHttpReq
     // Propagate Http/2 settings:
     delegate->http2Parameters = request.http2Configuration();
 
+    if (request.attribute(QNetworkRequest::ConnectionCacheExpiryTimeoutSecondsAttribute).isValid())
+        delegate->connectionCacheExpiryTimeoutSeconds = request.attribute(QNetworkRequest::ConnectionCacheExpiryTimeoutSecondsAttribute).toInt();
+
     // For the synchronous HTTP, this is the normal way the delegate gets deleted
     // For the asynchronous HTTP this is a safety measure, the delegate deletes itself when HTTP is finished
     QMetaObject::Connection threadFinishedConnection =
@@ -839,14 +880,12 @@ void QNetworkReplyHttpImplPrivate::postRequest(const QNetworkRequest &newHttpReq
         QObject::connect(delegate, SIGNAL(downloadFinished()),
                 q, SLOT(replyFinished()),
                 Qt::QueuedConnection);
-        QObject::connect(delegate, SIGNAL(downloadMetaData(QList<QPair<QByteArray,QByteArray> >,
-                                                           int, QString, bool,
-                                                           QSharedPointer<char>, qint64, qint64,
-                                                           bool)),
-                q, SLOT(replyDownloadMetaData(QList<QPair<QByteArray,QByteArray> >,
-                                              int, QString, bool,
-                                              QSharedPointer<char>, qint64, qint64, bool)),
-                Qt::QueuedConnection);
+        QObject::connect(delegate, &QHttpThreadDelegate::socketStartedConnecting,
+                q, &QNetworkReply::socketStartedConnecting, Qt::QueuedConnection);
+        QObject::connect(delegate, &QHttpThreadDelegate::requestSent,
+                q, &QNetworkReply::requestSent, Qt::QueuedConnection);
+        connect(delegate, &QHttpThreadDelegate::downloadMetaData, this,
+                &QNetworkReplyHttpImplPrivate::replyDownloadMetaData, Qt::QueuedConnection);
         QObject::connect(delegate, SIGNAL(downloadProgress(qint64,qint64)),
                 q, SLOT(replyDownloadProgressSlot(qint64,qint64)),
                 Qt::QueuedConnection);
@@ -899,14 +938,14 @@ void QNetworkReplyHttpImplPrivate::postRequest(const QNetworkRequest &newHttpReq
             delegate->httpRequest.setUploadByteDevice(forwardUploadDevice);
 
             // If the device in the user thread claims it has more data, keep the flow to HTTP thread going
-            QObject::connect(uploadByteDevice.data(), SIGNAL(readyRead()),
+            QObject::connect(uploadByteDevice.get(), SIGNAL(readyRead()),
                              q, SLOT(uploadByteDeviceReadyReadSlot()),
                              Qt::QueuedConnection);
 
             // From user thread to http thread:
             QObject::connect(q, SIGNAL(haveUploadData(qint64,QByteArray,bool,qint64)),
                              forwardUploadDevice, SLOT(haveDataSlot(qint64,QByteArray,bool,qint64)), Qt::QueuedConnection);
-            QObject::connect(uploadByteDevice.data(), SIGNAL(readyRead()),
+            QObject::connect(uploadByteDevice.get(), SIGNAL(readyRead()),
                              forwardUploadDevice, SIGNAL(readyRead()),
                              Qt::QueuedConnection);
 
@@ -929,7 +968,7 @@ void QNetworkReplyHttpImplPrivate::postRequest(const QNetworkRequest &newHttpReq
             // use the uploadByteDevice provided to us by the QNetworkReplyImpl.
             // The code that is in start() makes sure it is safe to use from a thread
             // since it only wraps a QRingBuffer
-            delegate->httpRequest.setUploadByteDevice(uploadByteDevice.data());
+            delegate->httpRequest.setUploadByteDevice(uploadByteDevice.get());
         }
     }
 
@@ -946,30 +985,20 @@ void QNetworkReplyHttpImplPrivate::postRequest(const QNetworkRequest &newHttpReq
     if (synchronous) {
         emit q->startHttpRequestSynchronously(); // This one is BlockingQueuedConnection, so it will return when all work is done
 
-        if (delegate->incomingErrorCode != QNetworkReply::NoError) {
-            replyDownloadMetaData
-                    (delegate->incomingHeaders,
-                     delegate->incomingStatusCode,
-                     delegate->incomingReasonPhrase,
-                     delegate->isPipeliningUsed,
-                     QSharedPointer<char>(),
-                     delegate->incomingContentLength,
-                     delegate->removedContentLength,
-                     delegate->isHttp2Used);
-            replyDownloadData(delegate->synchronousDownloadData);
+        replyDownloadMetaData
+                (delegate->incomingHeaders,
+                    delegate->incomingStatusCode,
+                    delegate->incomingReasonPhrase,
+                    delegate->isPipeliningUsed,
+                    QSharedPointer<char>(),
+                    delegate->incomingContentLength,
+                    delegate->removedContentLength,
+                    delegate->isHttp2Used,
+                    delegate->isCompressed);
+        replyDownloadData(delegate->synchronousDownloadData);
+
+        if (delegate->incomingErrorCode != QNetworkReply::NoError)
             httpError(delegate->incomingErrorCode, delegate->incomingErrorDetail);
-        } else {
-            replyDownloadMetaData
-                    (delegate->incomingHeaders,
-                     delegate->incomingStatusCode,
-                     delegate->incomingReasonPhrase,
-                     delegate->isPipeliningUsed,
-                     QSharedPointer<char>(),
-                     delegate->incomingContentLength,
-                     delegate->removedContentLength,
-                     delegate->isHttp2Used);
-            replyDownloadData(delegate->synchronousDownloadData);
-        }
 
         thread->quit();
         thread->wait(QDeadlineTimer(5000));
@@ -1040,23 +1069,79 @@ void QNetworkReplyHttpImplPrivate::replyDownloadData(QByteArray d)
     if (!q->isOpen())
         return;
 
+    // cache this, we need it later and it's invalidated when dealing with compressed data
+    auto dataSize = d.size();
+    // Grab this to compare later (only relevant for compressed data) in case none of the data
+    // will be propagated to the user
+    const qint64 previousBytesDownloaded = bytesDownloaded;
+
     if (cacheEnabled && isCachingAllowed() && !cacheSaveDevice)
         initCacheSaveDevice();
 
+    if (decompressHelper.isValid()) {
+        qint64 uncompressedBefore = -1;
+        if (decompressHelper.isCountingBytes())
+            uncompressedBefore = decompressHelper.uncompressedSize();
+
+        decompressHelper.feed(std::move(d));
+
+        if (!decompressHelper.isValid()) {
+            error(QNetworkReplyImpl::NetworkError::UnknownContentError,
+                  QCoreApplication::translate("QHttp", "Decompression failed: %1")
+                          .arg(decompressHelper.errorString()));
+            return;
+        }
+
+        if (!isHttpRedirectResponse()) {
+            if (decompressHelper.isCountingBytes())
+                bytesDownloaded += (decompressHelper.uncompressedSize() - uncompressedBefore);
+            setupTransferTimeout();
+        }
+
+        if (synchronous) {
+            d = QByteArray();
+            const qsizetype increments = 16 * 1024;
+            qint64 bytesRead = 0;
+            while (decompressHelper.hasData()) {
+                quint64 nextSize = quint64(d.size()) + quint64(increments);
+                if (nextSize > quint64(std::numeric_limits<QByteArray::size_type>::max())) {
+                    error(QNetworkReplyImpl::NetworkError::UnknownContentError,
+                          QCoreApplication::translate("QHttp",
+                                                      "Data downloaded is too large to store"));
+                    return;
+                }
+                d.resize(nextSize);
+                bytesRead += decompressHelper.read(d.data() + bytesRead, increments);
+                if (!decompressHelper.isValid()) {
+                    error(QNetworkReplyImpl::NetworkError::UnknownContentError,
+                          QCoreApplication::translate("QHttp", "Decompression failed: %1")
+                                  .arg(decompressHelper.errorString()));
+                    return;
+                }
+            }
+            d.resize(bytesRead);
+            // we're synchronous so we're not calling this function again; reset the decompressHelper
+            decompressHelper.clear();
+        }
+    }
+
     // This is going to look a little strange. When downloading data while a
     // HTTP redirect is happening (and enabled), we write the redirect
     // response to the cache. However, we do not append it to our internal
     // buffer as that will contain the response data only for the final
     // response
-    if (cacheSaveDevice)
+    // Note: For compressed data this is done in readData()
+    if (cacheSaveDevice && !decompressHelper.isValid()) {
         cacheSaveDevice->write(d);
+    }
 
-    if (!isHttpRedirectResponse()) {
+    // if decompressHelper is valid then we have compressed data, and this is handled above
+    if (!decompressHelper.isValid() && !isHttpRedirectResponse()) {
         buffer.append(d);
-        bytesDownloaded += d.size();
+        bytesDownloaded += dataSize;
         setupTransferTimeout();
     }
-    bytesBuffered += d.size();
+    bytesBuffered += dataSize;
 
     int pendingSignals = pendingDownloadDataEmissions->fetchAndSubAcquire(1) - 1;
     if (pendingSignals > 0) {
@@ -1070,17 +1155,25 @@ void QNetworkReplyHttpImplPrivate::replyDownloadData(QByteArray d)
     if (isHttpRedirectResponse())
         return;
 
+    // This can occur when downloading compressed data as some of the data may be the content
+    // encoding's header. Don't emit anything for this.
+    if (previousBytesDownloaded == bytesDownloaded) {
+        if (readBufferMaxSize)
+            emit q->readBufferFreed(dataSize);
+        return;
+    }
+
     QVariant totalSize = cookedHeaders.value(QNetworkRequest::ContentLengthHeader);
 
     emit q->readyRead();
-    // emit readyRead before downloadProgress incase this will cause events to be
+    // emit readyRead before downloadProgress in case this will cause events to be
     // processed and we get into a recursive call (as in QProgressDialog).
-    if (downloadProgressSignalChoke.elapsed() >= progressSignalInterval) {
+    if (downloadProgressSignalChoke.elapsed() >= progressSignalInterval
+        && (!decompressHelper.isValid() || decompressHelper.isCountingBytes())) {
         downloadProgressSignalChoke.restart();
         emit q->downloadProgress(bytesDownloaded,
                              totalSize.isNull() ? Q_INT64_C(-1) : totalSize.toLongLong());
     }
-
 }
 
 void QNetworkReplyHttpImplPrivate::replyFinished()
@@ -1207,6 +1300,7 @@ void QNetworkReplyHttpImplPrivate::followRedirect()
     Q_Q(QNetworkReplyHttpImpl);
     Q_ASSERT(managerPrivate);
 
+    decompressHelper.clear();
     rawHeaders.clear();
     cookedHeaders.clear();
 
@@ -1242,7 +1336,7 @@ void QNetworkReplyHttpImplPrivate::replyDownloadMetaData(const QList<QPair<QByte
                                                          QSharedPointer<char> db,
                                                          qint64 contentLength,
                                                          qint64 removedContentLength,
-                                                         bool h2Used)
+                                                         bool h2Used, bool isCompressed)
 {
     Q_Q(QNetworkReplyHttpImpl);
     Q_UNUSED(contentLength);
@@ -1283,6 +1377,22 @@ void QNetworkReplyHttpImplPrivate::replyDownloadMetaData(const QList<QPair<QByte
         if (it->first.toLower() == "location")
             value.clear();
 
+        if (isCompressed && !decompressHelper.isValid()
+            && it->first.compare("content-encoding", Qt::CaseInsensitive) == 0) {
+
+            if (!synchronous) // with synchronous all the data is expected to be handled at once
+                decompressHelper.setCountingBytesEnabled(true);
+
+            if (!decompressHelper.setEncoding(it->second)) {
+                error(QNetworkReplyImpl::NetworkError::UnknownContentError,
+                      QCoreApplication::translate("QHttp", "Failed to initialize decompression: %1")
+                              .arg(decompressHelper.errorString()));
+                return;
+            }
+            decompressHelper.setDecompressedSafetyCheckThreshold(
+                    request.decompressedSafetyCheckThreshold());
+        }
+
         if (!value.isEmpty()) {
             // Why are we appending values for headers which are already
             // present?
@@ -1385,7 +1495,7 @@ void QNetworkReplyHttpImplPrivate::replyDownloadProgressSlot(qint64 bytesReceive
     downloadBufferCurrentSize = bytesReceived;
 
     // Only emit readyRead when actual data is there
-    // emit readyRead before downloadProgress incase this will cause events to be
+    // emit readyRead before downloadProgress in case this will cause events to be
     // processed and we get into a recursive call (as in QProgressDialog).
     if (bytesDownloaded > 0)
         emit q->readyRead();
@@ -1802,7 +1912,7 @@ void QNetworkReplyHttpImplPrivate::_q_cacheLoadReadyRead()
 
     QVariant totalSize = cookedHeaders.value(QNetworkRequest::ContentLengthHeader);
 
-    // emit readyRead before downloadProgress incase this will cause events to be
+    // emit readyRead before downloadProgress in case this will cause events to be
     // processed and we get into a recursive call (as in QProgressDialog).
 
     if (!(isHttpRedirectResponse())) {
@@ -1879,7 +1989,7 @@ void QNetworkReplyHttpImplPrivate::_q_bufferOutgoingData()
 
     if (!outgoingDataBuffer) {
         // first call, create our buffer
-        outgoingDataBuffer = QSharedPointer<QRingBuffer>::create();
+        outgoingDataBuffer = std::make_shared<QRingBuffer>();
 
         QObject::connect(outgoingData, SIGNAL(readyRead()), q, SLOT(_q_bufferOutgoingData()));
         QObject::connect(outgoingData, SIGNAL(readChannelFinished()), q, SLOT(_q_bufferOutgoingDataFinished()));
@@ -1979,10 +2089,10 @@ QNonContiguousByteDevice* QNetworkReplyHttpImplPrivate::createUploadByteDevice()
 
     // We want signal emissions only for normal asynchronous uploads
     if (!synchronous)
-        QObject::connect(uploadByteDevice.data(), SIGNAL(readProgress(qint64,qint64)),
+        QObject::connect(uploadByteDevice.get(), SIGNAL(readProgress(qint64,qint64)),
                          q, SLOT(emitReplyUploadProgress(qint64,qint64)));
 
-    return uploadByteDevice.data();
+    return uploadByteDevice.get();
 }
 
 void QNetworkReplyHttpImplPrivate::_q_finished()
@@ -2001,9 +2111,12 @@ void QNetworkReplyHttpImplPrivate::finished()
 
     QVariant totalSize = cookedHeaders.value(QNetworkRequest::ContentLengthHeader);
 
-    // if we don't know the total size of or we received everything save the cache
-    if (totalSize.isNull() || totalSize == -1 || bytesDownloaded == totalSize)
+    // if we don't know the total size of or we received everything save the cache.
+    // If the data is compressed then this is done in readData()
+    if ((totalSize.isNull() || totalSize == -1 || bytesDownloaded == totalSize)
+        && !decompressHelper.isValid()) {
         completeCacheSave();
+    }
 
     // We check for errorCode too as in case of SSL handshake failure, we still
     // get the HTTP redirect status code (301, 303 etc)
@@ -2041,6 +2154,9 @@ void QNetworkReplyHttpImplPrivate::error(QNetworkReplyImpl::NetworkError code, c
         return;
     }
 
+    if (decompressHelper.isValid())
+        decompressHelper.clear(); // Just get rid of any data that might be stored
+
     errorCode = code;
     q->setErrorString(errorMessage);
 
