@@ -375,8 +375,6 @@ bool QMYSQLResultPrivate::bindInValues()
 
         char *field = bind->buffer_length ? new char[bind->buffer_length + 1]{} : nullptr;
         bind->buffer = f.outField = field;
-        if (qIsTimeOrDate(fieldInfo->type))
-            new (field) MYSQL_TIME;
 
         ++i;
     }
@@ -427,11 +425,9 @@ void QMYSQLResult::cleanup()
         d->meta = 0;
     }
 
-    for (const auto &field : qAsConst(d->fields)) {
-        if (qIsTimeOrDate(field.myField->type))
-            reinterpret_cast<MYSQL_TIME *>(field.outField)->~MYSQL_TIME();
-        delete[] field.outField;
-    }
+    int i;
+    for (i = 0; i < d->fields.count(); ++i)
+        delete[] d->fields[i].outField;
 
     if (d->outBinds) {
         delete[] d->outBinds;
@@ -1553,3 +1549,5 @@ bool QMYSQLDriver::isIdentifierEscaped(const QString &identifier, IdentifierType
 }
 
 QT_END_NAMESPACE
+
+#include "moc_qsql_mysql_p.cpp"
