@@ -99,9 +99,9 @@ static inline QString msgErrorSettingEllipticCurves(const QString &why)
     return QSslSocket::tr("Error when setting the elliptic curves (%1)").arg(why);
 }
 
-long QSslContext::setupOpenSslOptions(QSsl::SslProtocol protocol, QSsl::SslOptions sslOptions)
+qssloptions QSslContext::setupOpenSslOptions(QSsl::SslProtocol protocol, QSsl::SslOptions sslOptions)
 {
-    long options;
+    qssloptions options;
     switch (protocol) {
 QT_WARNING_PUSH
 QT_WARNING_DISABLE_DEPRECATED
@@ -512,7 +512,7 @@ QT_WARNING_POP
     }
 
     // Enable bug workarounds.
-    const long options = setupOpenSslOptions(configuration.protocol(), configuration.d->sslOptions);
+    const qssloptions options = setupOpenSslOptions(configuration.protocol(), configuration.d->sslOptions);
     q_SSL_CTX_set_options(sslContext->ctx, options);
 
     // Tell OpenSSL to release memory early
