@@ -146,16 +146,39 @@ bool QDialogPrivate::canBeNativeDialog() const
 /*!
     \internal
 
-    Properly hides dialog and sets the \a resultCode.
+    Properly closes dialog and sets the \a resultCode.
  */
-void QDialogPrivate::hide(int resultCode)
+void QDialogPrivate::close(int resultCode)
 {
     Q_Q(QDialog);
 
     q->setResult(resultCode);
-    q->hide();
 
-    close_helper(QWidgetPrivate::CloseNoEvent);
+    if (!data.is_closing) {
+        // Until Qt 6.3 we didn't close dialogs, so they didn't receive a QCloseEvent.
+        // It is likely that subclasses implement closeEvent and handle them as rejection
+        // (like QMessageBox and QProgressDialog do), so eat those events.
+        struct CloseEventEater : QObject
+        {
+            using QObject::QObject;
+        protected:
+            bool eventFilter(QObject *o, QEvent *e) override
+            {
+                if (e->type() == QEvent::Close)
+                    return true;
+                return QObject::eventFilter(o, e);
+            }
+        } closeEventEater;
+        q->installEventFilter(&closeEventEater);
+        QWidgetPrivate::close();
+    } else {
+        // If the close was initiated outside of QDialog we will end up
+        // here via QDialog::closeEvent calling reject(), in which case
+        // we need to hide the dialog to ensure QDialog::closeEvent does
+        // not ignore the close event. FIXME: Why is QDialog doing this?
+        q->hide();
+    }
+
     resetModalitySetByOpen();
 }
 
@@ -357,8 +380,7 @@ void QDialogPrivate::deletePlatformHelper()
 
     \snippet dialogs/dialogs.cpp 0
 
-    \sa QDialogButtonBox, QTabWidget, QWidget, QProgressDialog,
-        {fowler}{GUI Design Handbook: Dialogs, Standard}, {Extension Example},
+    \sa QDialogButtonBox, QTabWidget, QWidget, QProgressDialog, {Extension Example},
         {Standard Dialogs Example}
 */
 
@@ -623,7 +645,7 @@ int QDialog::exec()
   As with QWidget::close(), done() deletes the dialog if the
   Qt::WA_DeleteOnClose flag is set. If the dialog is the application's
   main widget, the application terminates. If the dialog is the
-  last window closed, the QApplication::lastWindowClosed() signal is
+  last window closed, the QGuiApplication::lastWindowClosed() signal is
   emitted.
 
   \sa accept(), reject(), QApplication::activeWindow(), QCoreApplication::quit()
@@ -632,7 +654,7 @@ int QDialog::exec()
 void QDialog::done(int r)
 {
     Q_D(QDialog);
-    d->hide(r);
+    d->close(r);
     d->finalize(r, r);
 }
 
@@ -783,42 +805,47 @@ void QDialog::setVisible(bool visible)
             return;
 
         QWidget::setVisible(visible);
-        QWidget *fw = window()->focusWidget();
-        if (!fw)
-            fw = this;
 
-        /*
-          The following block is to handle a special case, and does not
-          really follow propper logic in concern of autoDefault and TAB
-          order. However, it's here to ease usage for the users. If a
-          dialog has a default QPushButton, and first widget in the TAB
-          order also is a QPushButton, then we give focus to the main
-          default QPushButton. This simplifies code for the developers,
-          and actually catches most cases... If not, then they simply
-          have to use [widget*]->setFocus() themselves...
-        */
+        // Window activation might be prevented. We can't test isActiveWindow here,
+        // as the window will be activated asynchronously by the window manager.
+        if (!testAttribute(Qt::WA_ShowWithoutActivating)) {
+            QWidget *fw = window()->focusWidget();
+            if (!fw)
+                fw = this;
+
+            /*
+            The following block is to handle a special case, and does not
+            really follow proper logic in concern of autoDefault and TAB
+            order. However, it's here to ease usage for the users. If a
+            dialog has a default QPushButton, and first widget in the TAB
+            order also is a QPushButton, then we give focus to the main
+            default QPushButton. This simplifies code for the developers,
+            and actually catches most cases... If not, then they simply
+            have to use [widget*]->setFocus() themselves...
+            */
 #if QT_CONFIG(pushbutton)
-        if (d->mainDef && fw->focusPolicy() == Qt::NoFocus) {
-            QWidget *first = fw;
-            while ((first = first->nextInFocusChain()) != fw && first->focusPolicy() == Qt::NoFocus)
-                ;
-            if (first != d->mainDef && qobject_cast<QPushButton*>(first))
-                d->mainDef->setFocus();
-        }
-        if (!d->mainDef && isWindow()) {
-            QWidget *w = fw;
-            while ((w = w->nextInFocusChain()) != fw) {
-                QPushButton *pb = qobject_cast<QPushButton *>(w);
-                if (pb && pb->autoDefault() && pb->focusPolicy() != Qt::NoFocus) {
-                    pb->setDefault(true);
-                    break;
+            if (d->mainDef && fw->focusPolicy() == Qt::NoFocus) {
+                QWidget *first = fw;
+                while ((first = first->nextInFocusChain()) != fw && first->focusPolicy() == Qt::NoFocus)
+                    ;
+                if (first != d->mainDef && qobject_cast<QPushButton*>(first))
+                    d->mainDef->setFocus();
+            }
+            if (!d->mainDef && isWindow()) {
+                QWidget *w = fw;
+                while ((w = w->nextInFocusChain()) != fw) {
+                    QPushButton *pb = qobject_cast<QPushButton *>(w);
+                    if (pb && pb->autoDefault() && pb->focusPolicy() != Qt::NoFocus) {
+                        pb->setDefault(true);
+                        break;
+                    }
                 }
             }
-        }
 #endif
-        if (fw && !fw->hasFocus()) {
-            QFocusEvent e(QEvent::FocusIn, Qt::TabFocusReason);
-            QCoreApplication::sendEvent(fw, &e);
+            if (fw && !fw->hasFocus()) {
+                QFocusEvent e(QEvent::FocusIn, Qt::TabFocusReason);
+                QCoreApplication::sendEvent(fw, &e);
+            }
         }
 
 #ifndef QT_NO_ACCESSIBILITY
