@@ -78,10 +78,13 @@ public:
             // explicitly and not because the parent window in this case made it hidden.
             // In which case do not automatically show the widget when the parent
             // window is shown.
+            const bool wasExplicitShowHide = widget->testAttribute(Qt::WA_WState_ExplicitShowHide);
             const bool wasHidden = widget->testAttribute(Qt::WA_WState_Hidden);
             QWidgetPrivate::get(widget)->setVisible(visible);
-            if (!wasHidden)
-                widget->setAttribute(Qt::WA_WState_ExplicitShowHide, false);
+            if (wasExplicitShowHide) {
+                widget->setAttribute(Qt::WA_WState_ExplicitShowHide, wasExplicitShowHide);
+                widget->setAttribute(Qt::WA_WState_Hidden, wasHidden);
+            }
         } else {
             QWindowPrivate::setVisible(visible);
         }
@@ -115,6 +118,9 @@ public:
         if (QWidget *widget = q->widget())
             QWidgetPrivate::get(widget)->updateContentsRect();
     }
+
+    bool participatesInLastWindowClosed() const override;
+    bool treatAsVisible() const override;
 };
 
 QRectF QWidgetWindowPrivate::closestAcceptableGeometry(const QRectF &rect) const
@@ -225,6 +231,7 @@ static inline bool shouldBePropagatedToWidget(QEvent *event)
     case QEvent::ChildAdded:
     case QEvent::ChildRemoved:
     case QEvent::Paint:
+    case QEvent::Close: // Propagated manually in closeEvent
         return false;
     default:
         return true;
@@ -236,24 +243,7 @@ bool QWidgetWindow::event(QEvent *event)
     if (!m_widget)
         return QWindow::event(event);
 
-    if (m_widget->testAttribute(Qt::WA_DontShowOnScreen)) {
-        // \a event is uninteresting for QWidgetWindow, the event was probably
-        // generated before WA_DontShowOnScreen was set
-        if (!shouldBePropagatedToWidget(event))
-            return true;
-        return QCoreApplication::forwardEvent(m_widget, event);
-    }
-
     switch (event->type()) {
-    case QEvent::Close: {
-        // The widget might be deleted in the close event handler.
-        QPointer<QObject> guard = this;
-        handleCloseEvent(static_cast<QCloseEvent *>(event));
-        if (guard)
-            QWindow::event(event);
-        return true;
-    }
-
     case QEvent::Enter:
     case QEvent::Leave:
         handleEnterLeaveEvent(event);
@@ -397,14 +387,13 @@ QPointer<QWidget> qt_last_mouse_receiver = nullptr;
 
 void QWidgetWindow::handleEnterLeaveEvent(QEvent *event)
 {
-#if !defined(Q_OS_MACOS) && !defined(Q_OS_IOS) // Cocoa tracks popups
     // Ignore all enter/leave events from QPA if we are not on the first-level context menu.
     // This prevents duplicated events on most platforms. Fake events will be delivered in
     // QWidgetWindow::handleMouseEvent(QMouseEvent *). Make an exception whether the widget
     // is already under mouse - let the mouse leave.
     if (QApplicationPrivate::inPopupMode() && m_widget != QApplication::activePopupWidget() && !m_widget->underMouse())
         return;
-#endif
+
     if (event->type() == QEvent::Leave) {
         QWidget *enter = nullptr;
         // Check from window system event queue if the next queued enter targets a window
@@ -413,7 +402,7 @@ void QWidgetWindow::handleEnterLeaveEvent(QEvent *event)
         QWindowSystemInterfacePrivate::EnterEvent *systemEvent =
             static_cast<QWindowSystemInterfacePrivate::EnterEvent *>
             (QWindowSystemInterfacePrivate::peekWindowSystemEvent(QWindowSystemInterfacePrivate::Enter));
-        const QPointF globalPosF = systemEvent ? systemEvent->globalPos : QGuiApplicationPrivate::lastCursorPosition;
+        const QPointF globalPosF = systemEvent ? systemEvent->globalPos : QPointF(QGuiApplicationPrivate::lastCursorPosition);
         if (systemEvent) {
             if (QWidgetWindow *enterWindow = qobject_cast<QWidgetWindow *>(systemEvent->enter))
             {
@@ -545,7 +534,6 @@ void QWidgetWindow::handleMouseEvent(QMouseEvent *event)
             if (receiver != activePopupWidget)
                 widgetPos = receiver->mapFromGlobal(event->globalPosition().toPoint());
 
-#if !defined(Q_OS_MACOS) && !defined(Q_OS_IOS) // Cocoa tracks popups
             const bool reallyUnderMouse = activePopupWidget->rect().contains(mapped);
             const bool underMouse = activePopupWidget->underMouse();
             if (underMouse != reallyUnderMouse) {
@@ -563,7 +551,7 @@ void QWidgetWindow::handleMouseEvent(QMouseEvent *event)
                     receiver = activePopupWidget;
                 }
             }
-#endif
+
             if ((event->type() != QEvent::MouseButtonPress) || !(QMutableSinglePointEvent::from(event)->isDoubleClick())) {
                 // if the widget that was pressed is gone, then deliver move events without buttons
                 const auto buttons = event->type() == QEvent::MouseMove && qt_popup_down_closed
@@ -629,7 +617,8 @@ void QWidgetWindow::handleMouseEvent(QMouseEvent *event)
                 receiver = qt_button_down;
             else if (popupChild)
                 receiver = popupChild;
-            QContextMenuEvent e(QContextMenuEvent::Mouse, mapped, event->globalPosition().toPoint(), event->modifiers());
+            const QPoint localPos = receiver->mapFromGlobal(event->globalPosition().toPoint());
+            QContextMenuEvent e(QContextMenuEvent::Mouse, localPos, event->globalPosition().toPoint(), event->modifiers());
             QApplication::forwardEvent(receiver, &e, event);
         }
 #else
@@ -698,9 +687,7 @@ void QWidgetWindow::handleTouchEvent(QTouchEvent *event)
         // events instead, which QWidgetWindow::handleMouseEvent will forward correctly:
         event->ignore();
     } else {
-        event->setAccepted(QApplicationPrivate::translateRawTouchEvent(m_widget, event->pointingDevice(),
-                                                                       const_cast<QList<QEventPoint> &>(event->points()),
-                                                                       event->timestamp()));
+        event->setAccepted(QApplicationPrivate::translateRawTouchEvent(m_widget, event));
     }
 }
 
@@ -725,6 +712,9 @@ bool QWidgetWindow::updateSize()
     bool changed = false;
     if (m_widget->testAttribute(Qt::WA_OutsideWSRange))
         return changed;
+    if (m_widget->testAttribute(Qt::WA_DontShowOnScreen))
+        return changed;
+
     if (m_widget->data->crect.size() != geometry().size()) {
         changed = true;
         m_widget->data->crect.setSize(geometry().size());
@@ -800,6 +790,8 @@ void QWidgetWindow::handleMoveEvent(QMoveEvent *event)
 {
     if (m_widget->testAttribute(Qt::WA_OutsideWSRange))
         return;
+    if (m_widget->testAttribute(Qt::WA_DontShowOnScreen))
+        return;
 
     auto oldPosition = m_widget->data->crect.topLeft();
     auto newPosition = geometry().topLeft();
@@ -840,10 +832,38 @@ void QWidgetWindow::handleResizeEvent(QResizeEvent *event)
     }
 }
 
-void QWidgetWindow::handleCloseEvent(QCloseEvent *event)
+void QWidgetWindow::closeEvent(QCloseEvent *event)
 {
-    bool is_closing = m_widget->d_func()->close_helper(QWidgetPrivate::CloseWithSpontaneousEvent);
-    event->setAccepted(is_closing);
+    Q_D(QWidgetWindow);
+    bool accepted = m_widget->d_func()->handleClose(d->inClose ? QWidgetPrivate::CloseWithEvent
+                                                                  : QWidgetPrivate::CloseWithSpontaneousEvent);
+    event->setAccepted(accepted);
+}
+
+bool QWidgetWindowPrivate::participatesInLastWindowClosed() const
+{
+    Q_Q(const QWidgetWindow);
+
+    // For historical reasons WA_QuitOnClose has been closely tied
+    // to the lastWindowClosed signal, since the default behavior
+    // is to quit the application after emitting lastWindowClosed.
+    // ### Qt 7: Rename this attribute, or decouple behavior.
+    if (!q->widget()->testAttribute(Qt::WA_QuitOnClose))
+        return false;
+
+    return QWindowPrivate::participatesInLastWindowClosed();
+}
+
+bool QWidgetWindowPrivate::treatAsVisible() const
+{
+    Q_Q(const QWidgetWindow);
+
+    // Widget windows may have Qt::WA_DontShowOnScreen, in which case the
+    // QQWidget will be visible, but the corresponding QWindow will not.
+    // Since the lastWindowClosed logic relies on checking whether the
+    // closed window was visible, and if there are any remaining visible
+    // windows, we need to reflect the QWidget state, not the QWindow one.
+    return q->widget()->isVisible();
 }
 
 #if QT_CONFIG(wheelevent)
@@ -921,8 +941,9 @@ void QWidgetWindow::handleDragMoveEvent(QDragMoveEvent *event)
         event->ignore();
         if (m_dragTarget) { // Send DragLeave to previous
             QDragLeaveEvent leaveEvent;
-            QGuiApplication::forwardEvent(m_dragTarget, &leaveEvent, event);
+            QWidget *dragTarget = m_dragTarget;
             m_dragTarget = nullptr;
+            QGuiApplication::forwardEvent(dragTarget, &leaveEvent, event);
         }
     } else {
         const QPoint mapped = widget->mapFromGlobal(m_widget->mapToGlobal(event->position().toPoint()));
@@ -936,8 +957,9 @@ void QWidgetWindow::handleDragMoveEvent(QDragMoveEvent *event)
         } else {
             if (m_dragTarget) { // Send DragLeave to previous
                 QDragLeaveEvent leaveEvent;
-                QGuiApplication::forwardEvent(m_dragTarget, &leaveEvent, event);
+                QWidget *dragTarget = m_dragTarget;
                 m_dragTarget = nullptr;
+                QGuiApplication::forwardEvent(dragTarget, &leaveEvent, event);
             }
             // widget might have been deleted when handling the leaveEvent
             if (widget) {
@@ -959,9 +981,11 @@ void QWidgetWindow::handleDragMoveEvent(QDragMoveEvent *event)
 
 void QWidgetWindow::handleDragLeaveEvent(QDragLeaveEvent *event)
 {
-    if (m_dragTarget)
-        QGuiApplication::forwardEvent(m_dragTarget, event);
-    m_dragTarget = nullptr;
+    if (m_dragTarget) {
+        QWidget *dragTarget = m_dragTarget;
+        m_dragTarget = nullptr;
+        QGuiApplication::forwardEvent(dragTarget, event);
+    }
 }
 
 void QWidgetWindow::handleDropEvent(QDropEvent *event)
@@ -973,16 +997,20 @@ void QWidgetWindow::handleDropEvent(QDropEvent *event)
     }
     const QPoint mapped = m_dragTarget->mapFromGlobal(m_widget->mapToGlobal(event->position().toPoint()));
     QDropEvent translated(mapped, event->possibleActions(), event->mimeData(), event->buttons(), event->modifiers());
-    QGuiApplication::forwardEvent(m_dragTarget, &translated, event);
+    QWidget *dragTarget = m_dragTarget;
+    m_dragTarget = nullptr;
+    QGuiApplication::forwardEvent(dragTarget, &translated, event);
     event->setAccepted(translated.isAccepted());
     event->setDropAction(translated.dropAction());
-    m_dragTarget = nullptr;
 }
 
 #endif // QT_CONFIG(draganddrop)
 
 void QWidgetWindow::handleExposeEvent(QExposeEvent *event)
 {
+    if (m_widget->testAttribute(Qt::WA_DontShowOnScreen))
+        return; // Ignore for widgets that fake exposure
+
     QWidgetPrivate *wPriv = m_widget->d_func();
     const bool exposed = isExposed();
 
@@ -1002,7 +1030,7 @@ void QWidgetWindow::handleExposeEvent(QExposeEvent *event)
             // If the window becomes not exposed...
             if (wPriv->childrenShownByExpose) {
                 // ... and child widgets was previously shown by the expose event - hide widgets again.
-                // This is a workaround, because sometimes when window is minimized programatically,
+                // This is a workaround, because sometimes when window is minimized programmatically,
                 // the QPA can notify that the window is exposed after changing window state to minimized
                 // and then, the QPA can send next expose event with null exposed region (not exposed).
                 wPriv->hideChildren(true);
@@ -1018,11 +1046,8 @@ void QWidgetWindow::handleExposeEvent(QExposeEvent *event)
         m_widget->setAttribute(Qt::WA_Mapped);
         for (QWidget *p = m_widget->parentWidget(); p && !p->testAttribute(Qt::WA_Mapped); p = p->parentWidget())
             p->setAttribute(Qt::WA_Mapped);
-QT_WARNING_PUSH
-QT_WARNING_DISABLE_DEPRECATED
-        if (!event->region().isNull())
-            wPriv->syncBackingStore(event->region());
-QT_WARNING_POP
+        if (!event->m_region.isNull())
+            wPriv->syncBackingStore(event->m_region);
     } else {
         m_widget->setAttribute(Qt::WA_Mapped, false);
     }
