@@ -26,7 +26,8 @@
 **
 ****************************************************************************/
 
-#include <QtTest/QtTest>
+#include <QTest>
+#include <QSignalSpy>
 #include <QtCore/QCoreApplication>
 #include <qdebug.h>
 #include "modelstotest.cpp"
@@ -125,6 +126,7 @@ private:
 tst_QItemModel::tst_QItemModel()
 {
     qRegisterMetaType<QAbstractItemModel::LayoutChangeHint>();
+    qRegisterMetaType<QList<QPersistentModelIndex>>();
 }
 
 void tst_QItemModel::init()
@@ -181,7 +183,7 @@ void tst_QItemModel::nonDestructiveBasicTest()
     currentModel->hasChildren(QModelIndex());
     currentModel->hasIndex(0, 0);
     currentModel->headerData(0, Qt::Horizontal);
-    currentModel->index(0,0), QModelIndex();
+    currentModel->index(0,0);
     currentModel->itemData(QModelIndex());
     QVariant cache;
     currentModel->match(QModelIndex(), -1, cache);
@@ -576,12 +578,12 @@ void tst_QItemModel::data()
                 alignment == Qt::AlignJustify);
     }
 
-    QVariant colorVariant = currentModel->data(currentModel->index(0,0), Qt::BackgroundColorRole);
+    QVariant colorVariant = currentModel->data(currentModel->index(0,0), Qt::BackgroundRole);
     if (colorVariant.isValid()) {
         QVERIFY(colorVariant.canConvert<QColor>());
     }
 
-    colorVariant = currentModel->data(currentModel->index(0,0), Qt::TextColorRole);
+    colorVariant = currentModel->data(currentModel->index(0,0), Qt::ForegroundRole);
     if (colorVariant.isValid()) {
         QVERIFY(colorVariant.canConvert<QColor>());
     }
