@@ -32,10 +32,12 @@
 #include <QNetworkAccessManager>
 #include <QNetworkReply>
 #include <QNetworkRequest>
-#include <QtTest/QtTest>
+#include <QTest>
 #include <QUrl>
-#include <QXmlDefaultHandler>
 #include <QXmlStreamReader>
+#include <QBuffer>
+#include <QStack>
+#include <QtGui/private/qzipreader_p.h>
 
 #include "qc14n.h"
 
@@ -44,6 +46,11 @@ Q_DECLARE_METATYPE(QXmlStreamReader::ReadElementTextBehaviour)
 static const char *const catalogFile = "XML-Test-Suite/xmlconf/finalCatalog.xml";
 static const int expectedRunCount = 1646;
 static const int expectedSkipCount = 532;
+static const char *const xmlTestsuiteDir = "XML-Test-Suite";
+static const char *const xmlconfDir = "XML-Test-Suite/xmlconf/";
+static const char *const xmlDatasetName = "xmltest";
+static const char *const updateFilesDir = "xmltest_updates";
+static const char *const destinationFolder = "/valid/sa/out/";
 
 static inline int best(int a, int b)
 {
@@ -65,6 +72,28 @@ static inline int best(int a, int b, int c)
     return qMin(qMin(a, b), c);
 }
 
+// copied from tst_qmake.cpp
+static void copyDir(const QString &sourceDirPath, const QString &targetDirPath)
+{
+    QDir currentDir;
+    QDirIterator dit(sourceDirPath, QDir::Dirs | QDir::NoDotAndDotDot | QDir::Hidden);
+    while (dit.hasNext()) {
+        dit.next();
+        const QString targetPath = targetDirPath + QLatin1Char('/') + dit.fileName();
+        currentDir.mkpath(targetPath);
+        copyDir(dit.filePath(), targetPath);
+    }
+
+    QDirIterator fit(sourceDirPath, QDir::Files | QDir::Hidden);
+    while (fit.hasNext()) {
+        fit.next();
+        const QString targetPath = targetDirPath + QLatin1Char('/') + fit.fileName();
+        QFile::remove(targetPath);  // allowed to fail
+        QFile src(fit.filePath());
+        QVERIFY2(src.copy(targetPath), qPrintable(src.errorString()));
+    }
+}
+
 template <typename C>
 const C sorted_by_name(C c) { // return by const value so we can feed directly into range-for loops below
     using T = typename C::value_type;
@@ -221,7 +250,7 @@ static QString documentElement(const QByteArray &document)
  *
  * See \l {http://www.w3.org/XML/Test/} {Extensible Markup Language (XML) Conformance Test Suites}
  */
-class TestSuiteHandler : public QXmlDefaultHandler
+class TestSuiteHandler
 {
 public:
     /**
@@ -240,8 +269,8 @@ public:
      */
     class MissedBaseline
     {
-        friend class QVector<MissedBaseline>;
-        MissedBaseline() {} // for QVector, don't use
+        friend class QList<MissedBaseline>;
+        MissedBaseline() {} // for QList, don't use
     public:
         MissedBaseline(const QString &aId,
                        const QByteArray &aExpected,
@@ -265,8 +294,8 @@ public:
         QByteArray  output;
     };
 
-    QVector<GeneralFailure> failures;
-    QVector<MissedBaseline> missedBaselines;
+    QList<GeneralFailure> failures;
+    QList<MissedBaseline> missedBaselines;
 
     /**
      * The count of how many tests that were run.
@@ -286,29 +315,33 @@ public:
         m_baseURI.push(baseURI);
     }
 
-    virtual bool characters(const QString &chars)
+    bool runTests(QFile *file)
     {
-        m_ch = chars;
-        return true;
+        QXmlStreamReader reader(file);
+        while (!reader.atEnd() && !reader.hasError()) {
+            reader.readNext();
+
+            if (reader.isStartElement() && !startElement(reader.attributes()))
+                return false;
+
+            if (reader.isEndElement() && !endElement(reader.name().toString()))
+                return false;
+        }
+        return !reader.hasError();
     }
 
-    virtual bool startElement(const QString &,
-                              const QString &,
-                              const QString &,
-                              const QXmlAttributes &atts)
+    bool startElement(const QXmlStreamAttributes &atts)
     {
         m_atts.push(atts);
-        const int i = atts.index(QLatin1String("xml:base"));
 
-        if(i != -1)
-            m_baseURI.push(m_baseURI.top().resolved(atts.value(i)));
+        const auto attr = atts.value(QLatin1String("xml:base"));
+        if (!attr.isEmpty())
+            m_baseURI.push(m_baseURI.top().resolved(attr.toString()));
 
         return true;
     }
 
-    virtual bool endElement(const QString &,
-                            const QString &localName,
-                            const QString &)
+    bool endElement(const QString &localName)
     {
         if(localName == QLatin1String("TEST"))
         {
@@ -329,19 +362,19 @@ public:
                 return true;
             }
 
-            const QString inputFilePath(m_baseURI.top().resolved(m_atts.top().value(QString(), QLatin1String("URI")))
-                                                                .toLocalFile());
-            const QString id(m_atts.top().value(QString(), QLatin1String("ID")));
-            const QString type(m_atts.top().value(QString(), QLatin1String("TYPE")));
+            const QString inputFilePath(
+                    m_baseURI.top()
+                            .resolved(
+                                    m_atts.top().value(QString(), QLatin1String("URI")).toString())
+                            .toLocalFile());
+            const QString id(m_atts.top().value(QString(), QLatin1String("ID")).toString());
+            const QString type(m_atts.top().value(QString(), QLatin1String("TYPE")).toString());
 
             QString expectedFilePath;
-            const int index = m_atts.top().index(QString(), QLatin1String("OUTPUT"));
 
-            if(index != -1)
-            {
-                expectedFilePath = m_baseURI.top().resolved(m_atts.top().value(QString(),
-                                                            QLatin1String("OUTPUT"))).toLocalFile();
-            }
+            const auto attr = m_atts.top().value(QString(), QLatin1String("OUTPUT"));
+            if (!attr.isEmpty())
+                expectedFilePath = m_baseURI.top().resolved(attr.toString()).toLocalFile();
 
             /* testcases.dtd: 'No parser should accept a "not-wf" testcase
              * unless it's a nonvalidating parser and the test contains
@@ -349,7 +382,7 @@ public:
              *
              * We also let this apply to "valid", "invalid" and "error" tests, although
              * I'm not fully sure this is correct. */
-            const QString ents(m_atts.top().value(QString(), QLatin1String("ENTITIES")));
+            const QString ents(m_atts.top().value(QString(), QLatin1String("ENTITIES")).toString());
             m_atts.pop();
 
             if(ents == QLatin1String("both") ||
@@ -393,8 +426,6 @@ public:
                 return true;
             }
 
-            QXmlStreamReader reader(&inputFile);
-
             /* See testcases.dtd which reads: 'Nonvalidating parsers
              * must also accept "invalid" testcases, but validating ones must reject them.' */
             if(type == QLatin1String("invalid") || type == QLatin1String("valid"))
@@ -455,8 +486,8 @@ public:
                 qFatal("The input catalog is invalid.");
                 return false;
             }
-        }
-        else if(localName == QLatin1String("TESTCASES") && m_atts.top().index(QLatin1String("xml:base")) != -1)
+        } else if (localName == QLatin1String("TESTCASES")
+                   && m_atts.top().hasAttribute(QLatin1String("xml:base")))
             m_baseURI.pop();
 
         m_atts.pop();
@@ -516,9 +547,8 @@ public:
     }
 
 private:
-    QStack<QXmlAttributes>  m_atts;
-    QString                 m_ch;
-    QStack<QUrl>            m_baseURI;
+    QStack<QXmlStreamAttributes> m_atts;
+    QStack<QUrl> m_baseURI;
 };
 QT_BEGIN_NAMESPACE
 Q_DECLARE_SHARED(TestSuiteHandler::MissedBaseline)
@@ -528,7 +558,7 @@ class tst_QXmlStream: public QObject
 {
     Q_OBJECT
 public:
-    tst_QXmlStream() : m_handler(QUrl::fromLocalFile(QFINDTESTDATA(catalogFile)))
+    tst_QXmlStream() : m_handler(QUrl::fromLocalFile(m_tempDir.filePath(catalogFile)))
     {
     }
 
@@ -559,9 +589,7 @@ private slots:
     void crashInUTF16Codec() const;
     void hasAttributeSignature() const;
     void hasAttribute() const;
-    void writeWithCodec() const;
     void writeWithUtf8Codec() const;
-    void writeWithUtf16Codec() const;
     void writeWithStandalone() const;
     void entitiesAndWhitespace_1() const;
     void entitiesAndWhitespace_2() const;
@@ -572,7 +600,6 @@ private slots:
     void checkCommentIndentation() const;
     void checkCommentIndentation_data() const;
     void crashInXmlStreamReader() const;
-    void write8bitCodec() const;
     void invalidStringCharacters_data() const;
     void invalidStringCharacters() const;
     void hasError() const;
@@ -580,28 +607,54 @@ private slots:
     void roundTrip() const;
     void roundTrip_data() const;
 
+    void entityExpansionLimit() const;
+
 private:
     static QByteArray readFile(const QString &filename);
 
+    QTemporaryDir m_tempDir;
     TestSuiteHandler m_handler;
 };
 
 void tst_QXmlStream::initTestCase()
 {
-    QFile file(QFINDTESTDATA(catalogFile));
+    // Due to license restrictions, we need to distribute part of the test
+    // suit as a zip archive. So we need to unzip it before running the tests,
+    // and also update some files there.
+    // We also need to remove the unzipped data during cleanup.
+
+    // On Android, we cannot unzip at the resource location, so we copy
+    // everything to a temporary directory first.
+    const QString XML_Test_Suite_dir = QFINDTESTDATA(xmlTestsuiteDir);
+    const QString XML_Test_Suite_destDir = m_tempDir.filePath(xmlTestsuiteDir);
+    copyDir(XML_Test_Suite_dir, XML_Test_Suite_destDir);
+
+
+    const QString filesDir(m_tempDir.filePath(xmlconfDir));
+    const QString fileName = filesDir + xmlDatasetName + ".zip";
+    QVERIFY(QFile::exists(fileName));
+    QZipReader reader(fileName);
+    QVERIFY(reader.isReadable());
+    QVERIFY(reader.extractAll(filesDir));
+    // update files
+    const auto files =
+            QDir(filesDir + updateFilesDir).entryInfoList(QDir::Files | QDir::NoDotAndDotDot);
+    for (const auto &fileInfo : files) {
+        const QString destinationPath =
+                filesDir + xmlDatasetName + destinationFolder + fileInfo.fileName();
+        QFile::remove(destinationPath); // copy will fail if file exists
+        QVERIFY(QFile::copy(fileInfo.filePath(), destinationPath));
+    }
+
+    QFile file(m_tempDir.filePath(catalogFile));
     QVERIFY2(file.open(QIODevice::ReadOnly),
              qPrintable(QString::fromLatin1("Failed to open the test suite catalog; %1").arg(file.fileName())));
 
-    QXmlInputSource source(&file);
-    QXmlSimpleReader reader;
-    reader.setContentHandler(&m_handler);
-
-    QVERIFY(reader.parse(&source, false));
+    QVERIFY(m_handler.runTests(&file));
 }
 
 void tst_QXmlStream::cleanupTestCase()
 {
-    QFile::remove(QLatin1String("test.xml"));
 }
 
 void tst_QXmlStream::reportFailures() const
@@ -702,7 +755,7 @@ QByteArray tst_QXmlStream::readFile(const QString &filename)
     QByteArray outarray;
     QTextStream writer(&outarray);
     // We always want UTF-8, and not what the system picks up.
-    writer.setCodec("UTF-8");
+    writer.setEncoding(QStringConverter::Utf8);
 
     while (!reader.atEnd()) {
         reader.readNext();
@@ -857,7 +910,7 @@ void tst_QXmlStream::addExtraNamespaceDeclarations()
 
 class EntityResolver : public QXmlStreamEntityResolver {
 public:
-    QString resolveUndeclaredEntity(const QString &name) {
+    QString resolveUndeclaredEntity(const QString &name) override {
         static int count = 0;
         return name.toUpper() + QString::number(++count);
     }
@@ -1259,66 +1312,16 @@ void tst_QXmlStream::hasAttribute() const
     QVERIFY(!reader.hasError());
 }
 
-
-void tst_QXmlStream::writeWithCodec() const
-{
-    QByteArray outarray;
-    QXmlStreamWriter writer(&outarray);
-    writer.setAutoFormatting(true);
-
-    QTextCodec *codec = QTextCodec::codecForName("ISO 8859-15");
-    QVERIFY(codec);
-    writer.setCodec(codec);
-
-    const char *latin2 = "h\xe9 h\xe9";
-    const QString string = codec->toUnicode(latin2);
-
-
-    writer.writeStartDocument("1.0");
-
-    writer.writeTextElement("foo", string);
-    writer.writeEndElement();
-    writer.writeEndDocument();
-
-    QVERIFY(outarray.contains(latin2));
-    QVERIFY(outarray.contains(codec->name()));
-}
-
 void tst_QXmlStream::writeWithUtf8Codec() const
 {
     QByteArray outarray;
     QXmlStreamWriter writer(&outarray);
 
-    QTextCodec *codec = QTextCodec::codecForMib(106); // utf-8
-    QVERIFY(codec);
-    writer.setCodec(codec);
-
     writer.writeStartDocument("1.0");
     static const char begin[] = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
     QVERIFY(outarray.startsWith(begin));
 }
 
-void tst_QXmlStream::writeWithUtf16Codec() const
-{
-    QByteArray outarray;
-    QXmlStreamWriter writer(&outarray);
-
-    QTextCodec *codec = QTextCodec::codecForMib(1014); // utf-16LE
-    QVERIFY(codec);
-    writer.setCodec(codec);
-
-    writer.writeStartDocument("1.0");
-    static const char begin[] = "<?xml version=\"1.0\" encoding=\"UTF-16";  // skip potential "LE" suffix
-    const int count = sizeof(begin) - 1;    // don't include 0 terminator
-    QByteArray begin_UTF16;
-    begin_UTF16.reserve(2*(count));
-    for (int i = 0; i < count; ++i) {
-        begin_UTF16.append(begin[i]);
-        begin_UTF16.append((char)'\0');
-    }
-    QVERIFY(outarray.startsWith(begin_UTF16));
-}
-
 void tst_QXmlStream::writeWithStandalone() const
 {
     {
@@ -1414,7 +1417,6 @@ void tst_QXmlStream::garbageInXMLPrologUTF8Explicitly() const
     QVERIFY(out.open(QIODevice::ReadWrite));
 
     QXmlStreamWriter writer (&out);
-    writer.setCodec("UTF-8");
     writer.writeStartDocument();
     writer.writeEmptyElement("Foo");
     writer.writeEndDocument();
@@ -1522,7 +1524,7 @@ void tst_QXmlStream::crashInXmlStreamReader() const
 class FakeBuffer : public QBuffer
 {
 protected:
-    qint64 writeData(const char *c, qint64 i)
+    qint64 writeData(const char *c, qint64 i) override
     {
         qint64 ai = qMin(m_capacity, i);
         m_capacity -= ai;
@@ -1603,43 +1605,6 @@ void tst_QXmlStream::hasError() const
 
 }
 
-void tst_QXmlStream::write8bitCodec() const
-{
-    QBuffer outBuffer;
-    QVERIFY(outBuffer.open(QIODevice::WriteOnly));
-    QXmlStreamWriter writer(&outBuffer);
-    writer.setAutoFormatting(false);
-
-    QTextCodec *codec = QTextCodec::codecForName("IBM500");
-    if (!codec) {
-        QSKIP("Encoding IBM500 not available.");
-    }
-    writer.setCodec(codec);
-
-    writer.writeStartDocument();
-    writer.writeStartElement("root");
-    writer.writeAttribute("attrib", "1");
-    writer.writeEndElement();
-    writer.writeEndDocument();
-    outBuffer.close();
-
-    // test 8 bit encoding
-    QByteArray values = outBuffer.data();
-    QVERIFY(values.size() > 1);
-    // check '<'
-    QCOMPARE(values[0] & 0x00FF, 0x4c);
-    // check '?'
-    QCOMPARE(values[1] & 0x00FF, 0x6F);
-
-    // convert the start of the XML
-    const QString expected = ("<?xml version=\"1.0\" encoding=\"IBM500\"?>");
-    QTextDecoder *decoder = codec->makeDecoder();
-    QVERIFY(decoder);
-    QString decodedText = decoder->toUnicode(values);
-    delete decoder;
-    QVERIFY(decodedText.startsWith(expected));
-}
-
 void tst_QXmlStream::invalidStringCharacters() const
 {
     // test scan in attributes
@@ -1756,6 +1721,46 @@ void tst_QXmlStream::roundTrip_data() const
         "</root>\n";
 }
 
+void tst_QXmlStream::entityExpansionLimit() const
+{
+    QString xml = QStringLiteral("<?xml version=\"1.0\"?>"
+                                 "<!DOCTYPE foo ["
+                                 "<!ENTITY a \"0123456789\" >"
+                                 "<!ENTITY b \"&a;&a;&a;&a;&a;&a;&a;&a;&a;&a;\" >"
+                                 "<!ENTITY c \"&b;&b;&b;&b;&b;&b;&b;&b;&b;&b;\" >"
+                                 "<!ENTITY d \"&c;&c;&c;&c;&c;&c;&c;&c;&c;&c;\" >"
+                                 "]>"
+                                 "<foo>&d;&d;&d;</foo>");
+    {
+        QXmlStreamReader reader(xml);
+        QCOMPARE(reader.entityExpansionLimit(), 4096);
+        do {
+            reader.readNext();
+        } while (!reader.atEnd());
+        QCOMPARE(reader.error(), QXmlStreamReader::NotWellFormedError);
+    }
+
+    // &d; expands to 10k characters, minus the 3 removed (&d;) means it should fail
+    // with a limit of 9996 chars and pass with 9997
+    {
+        QXmlStreamReader reader(xml);
+        reader.setEntityExpansionLimit(9996);
+        do {
+            reader.readNext();
+        } while (!reader.atEnd());
+
+        QCOMPARE(reader.error(), QXmlStreamReader::NotWellFormedError);
+    }
+    {
+        QXmlStreamReader reader(xml);
+        reader.setEntityExpansionLimit(9997);
+        do {
+            reader.readNext();
+        } while (!reader.atEnd());
+        QCOMPARE(reader.error(), QXmlStreamReader::NoError);
+    }
+}
+
 void tst_QXmlStream::roundTrip() const
 {
     QFETCH(QString, in);
